from multiprocessing_utils import cpus
from date_iterator import DateIterator, DateBinner
from persistent_ordered_dict import (
	PersistentOrderedDict, ProgressTracker, DuplicateKeyException, SharedProgressTracker,
	PersistentOrderedDictException, PersistentOrderedDictException,
	PersistentOrderedDictIntegrityException
)
from safe import (
	safe_min, safe_max, safe_lte, safe_lt, safe_gte, safe_gt
)
from tsv import UnicodeTsvReader, UnicodeTsvWriter
from file_utils import ls, file_empty, ensure_exists, ensure_removed
from managed_process import ManagedProcess, DONE, NOT_DONE, SUICIDE
#from selenium_crawler import SeleniumCrawler, uses_selenium
from string_alignment import (
	StringAligner, string_distance, string_align, 
	string_align_masks, string_align_path, alignment_score
)
from grouper import (
	chunk, group, flatten, lindex, rindex, indices, skip, IncrementingMap,
	rangify, ltrim, rtrim, trim, binify, inbin, get_fold, skipfirst
)
import patterns
from trace import trace, get_trace
from io_util import out
#from vectorize import Vectorizer
from js import json_get_fast
from dictionary import (
    invert_dict, dzip, merge_dicts, select, min_item,
    max_item, max_key, min_key, max_value, min_value
)
from id_generator import UniqueIdGenerator, get_id
from track_progress import progress, pc
from extrema import Max, Min
from sample import reservoir_sample, ReservoirSampler
from unigram_dictionary import UnigramDictionary
from token_map import TokenMap, SILENT, WARN, ERROR, UNK
from counter_sampler import CounterSampler
from crowdflower_results import CrowdflowerResults
import html
from mem_server import MemoryServer, mem_get, remember
from ensure_unicode import ensure_unicode



'''
The TokenMap provides a mapping from tokens (plain words) to integer ids,
as well as a map from ids back to tokens.
'''

import gzip


SILENT = 0
WARN = 1
ERROR = 2
UNK = 0

def ensure_str(s):
	'''
	Ensures that the string is encoded as str, not unicode
	'''
	try:
		return s.encode('utf8')
	except UnicodeDecodeError:
		return s


# TODO: enable loading and saving
class TokenMap(object):

	def __init__(self, on_unk=WARN, tokens=None):
		'''
		Create a new TokenMap.  Most common usage is to call this without
		any arguments.

		on_unk:	Controls the behavior when asked to provide the token_id
			for a token not found in the map.  Default is WARN, which 
			means returning 0 (which is id reserved for unknown tokens) 
			and then printing a warning to stout.  Choose from SILENT, 
			WARN, or ERROR.

		tokens:	List of strings corresponding to a map that should be 
			used.  The index of a token in the list is used as its ID.
			Not normally used, because TokenMap provides functions to 
			build the map easily from a corpus.  The first element in the 
			list should be 'UNK', becuase id 0 is reserved for unknown 
			tokens.  Not doing so is an error.
		'''

		# Validate on_unk
		if on_unk not in (SILENT, WARN, ERROR):
			raise ValueError(
				'on_unk must be one of token_map.SILENT, token_map.WARN, '
				'or token_map.ERROR.'
			)
		self.on_unk = on_unk

		# Initialize the token mapping
		if tokens is None:
			self.map = {'UNK':UNK}	# keys are tokens, values are ids
			self.tokens = ['UNK']	# entries are tokens, indices are ids

		# If an initial lexicon was provided, build the map from it
		else:
			if tokens[0] != 'UNK':
				raise ValueError(
					'tokens[0] must be "UNK" because ID 0 is reserved for '
					'unknown tokens.'
				)

			self.tokens = [ensure_str(t) for t in tokens]
			self.map = dict((t, idx) for idx, t in enumerate(self.tokens))


	def compact(self):
		'''
		Recreate the tokens list and mapping such that `None`s are 
		removed (which are holes left by calls to `remove()`.
		'''
		self.tokens = [t for t in self.tokens if t is not None]
		self.map = dict((t, idx) for idx, t in enumerate(self.tokens))


	def remove(self, token):
		token = ensure_str(token)
		idx = self.get_id(token)
		if idx == UNK:
			raise ValueError(
				'Cannot remove token %s because it does not ' 
				'exist or is reserved.' % str(token)
			)
		self.tokens[idx] = None


	def add(self, token):
		token = ensure_str(token)
		try:
			return self.map[token]
		except KeyError:
			next_id = len(self.tokens)
			self.map[token] = next_id
			self.tokens.append(token)
			return next_id


	def get_vocab_size(self):
		return len(self.tokens)


	def update(self, token_iterable):
		return [self.add(token) for token in token_iterable]


	def get_id(self, token):
		token = ensure_str(token)
		try:
			return self.map[token]

		# If that token isn't in the vocabulary, what should we do?
		# This depends on the setting for on_unk.  We can return the UNK
		# token silently, return the UNK token with a warning, or 
		# raise an error.
		except KeyError:
			if self.on_unk == SILENT:
				return self.map['UNK']  # i.e. return 0
			elif self.on_unk == WARN:
				print 'Warning, unrecognized token: %s' % token
				return self.map['UNK']  # i.e. return 0
			elif self.on_unk == ERROR:
				raise
			else:
				raise ValueError(
					'Unrecognized value for on_unk in TokenMap.'
				)


	def get_ids(self, token_iterable):
		return [self.get_id(token) for token in token_iterable]


	def get_token(self, idx):
		return self.tokens[idx]

	def get_tokens(self, idx_iterable):
		return [self.tokens[idx] for idx in idx_iterable]


	def __len__(self):
		return len(self.tokens)


	def save(self, filename):
		if filename.endswith('.gz'):
			f = gzip.open(filename, 'w')
		else:
			f = open(filename, 'w')

		for idx, token in enumerate(self.tokens):
			f.write(token + '\n')


	def load(self, filename):
		self.map = {}
		self.tokens = []

		if filename.endswith('.gz'):
			f = gzip.open(filename)
		else:
			f = gzip.open(filename)

		for idx, line in enumerate(f):
			token = line.strip()
			self.map[token] = idx
			self.tokens.append(token)




##!/usr/bin/env python
#
#'''
#This module enables crawling reddit.com to obtain the html representing
#users' profiles.  It is executable and takes one arguments.  To crawl
#users, run `./crawl_users.py users`.
#
#The very first time you run this file, you may need to create a tracker,
#which is a file that the crawler uses to keep track of which users
#have already been crawled.  To do that, run `./crawl_users.py setup`.
#
#The tracker is built from a plain text list of 
#active users, which should be listed, one user per line, inside the file 
#`active-users.txt`.  This file should be stored in DATA_DIR 
#(which is a path specified in the project settings in `SETTINGS.py`).
#'''
#
#
#import time
#from pyvirtualdisplay import Display
#from selenium import webdriver
#from selenium.webdriver.support.ui import WebDriverWait 
#from selenium.webdriver.support.select import Select
#from selenium.common.exceptions import (
#	TimeoutException, NoSuchElementException, WebDriverException, 
#	StaleElementReferenceException
#)
#
#
#def uses_selenium(f):
#	'''
#	This is a decorator, which is intended to be used to decorate methods
#	of the Crawler class (or subclasses), which contain a selenium.webdriver
#	instance as the attribute `driver`.
#
#	Ensures that the webdriver is closed after a the decorated function 
#	completes, or in case an exception occurs within it.
#	'''
#
#	def wrapped(self, *args, **kwargs):
#
#		# If needed, make and start a display
#		if self.use_display:
#			self.display = Display(visible=0, size=(1024, 768))
#			self.display.start()
#
#		# make a selenium webdriver, and a wait
#		self.driver = webdriver.Firefox()
#		self._wait = WebDriverWait(self.driver, self.WAIT_TIME)
#
#		# Run the function in a try block
#		try:
#			return_val = f(self, *args, **kwargs)
#
#		# Quit driver and stop the display in case of exception...
#		except:
#			self.driver.quit()
#			if self.use_display:
#				self.display.stop()
#			raise
#
#		# ...Or after function completes
#		self.driver.quit()
#		if self.use_display:
#			self.display.stop()
#
#		# Return the original return value
#		return return_val
#
#	return wrapped
#
#
#class SeleniumCrawler(object):
#
#	REQUEST_DELAY = 1.2
#	WAIT_TIME = 30
#
#	# These are added to the crawler's namespace for convevenievce
#	# This way they don't need to be imported from Selenium
#	Select = Select
#	TimeoutException = TimeoutException
#	NoSuchElementException = NoSuchElementException
#	WebDriverException = WebDriverException
#	StaleElementReferenceException = StaleElementReferenceException
#
#
#	def __init__(self, use_display=True):
#		self.use_display = use_display
#
#
#	def delay(self):
#		'''
#		Ensures that between any two calls to delay(), at least 
#		self.REQUEST_DELAY seconds will have elapsed.
#		'''
#		if hasattr(self, '_timer'):
#			elapsed = time.time() - self._timer
#			remaining = self.REQUEST_DELAY - elapsed
#			time.sleep(max(0, remaining))
#
#		self._timer = time.time()
#
#
#	def quit_webdriver(self):
#		self.driver.quit()
#
#
#	def get_page_source(self):
#		return self.driver.page_source.encode('utf8')
#
#
#	def wait(self, f, args=(), kwargs={}):
#		return self._wait.until(lambda driver: f(*args, **kwargs))
#
#
#class UnexpectedPageException(Exception):
#	pass
#
#
#
#if __name__ == '__main__':
#	
#	if USE_PYVIRTUALDISPLAY:
#		display = Display(visible=0, size=(1024, 768))
#		display.start()
#
#	# Run the appropriate function.
#	try:
#		dispatch()
#
#	# Afterwards, or if there is an exception, stop the display if needed.
#	except:
#		if USE_PYVIRTUALDISPLAY:
#			display.stop()
#		raise
#	else:
#		if USE_PYVIRTUALDISPLAY:
#			display.stop()



import random
import hashlib

class UniqueIdGenerator():

	INCREMENTING_SALT = 0

	# Keep a local random number generator so it can be seeded without
	# affecting the random module's global state
	RNG = random.Random()

	def __init__(
		self,
		length=16,
		deterministic=True,
		seed=None,
	):
		# Validation
		if seed is not None and not isinstance(seed, int):
			raise ValueError(
				'UniqueIdGenerator: seed must be an int or None.'
			)
		if length < 0 or length > 16 or not isinstance(length, int):
			raise ValueError(
				'UniqueIdGenerator: length must be an int '
				'between 0 and 32.'
			)

		# Set the seed
		if seed is not None:
			self.RNG.seed(seed)
			self.INCREMENTING_SALT = seed

		# Register state variables locally
		self.deterministic = deterministic
		self.length = length

	def get_id(self, string=''):
		# Use system time to make hash non-deterministic (if desired)
		if not self.deterministic:
			string += '%f' % self.RNG.random()
		# Return the first `self.length` number of hash characters
		return hashlib.sha1(string).hexdigest()[:self.length]

	def get_ids(self, int_or_strings):
		# Resolve whether an int or strings were provided for the hashing
		if isinstance(int_or_strings):
			strings = ['' for i in range(int_or_strings)]
		else:
			strings = int_or_strings
		# Return a bunch of ids
		return [self.get_id(string) for string in strings]


ID_GENERATOR = UniqueIdGenerator(deterministic=False)
def get_id():
	return ID_GENERATOR.get_id()



import re

class SENTINAL():
    pass

def json_get_fast(json_string, key, default=SENTINAL()):
    '''
    uses regex to pull out the simple value of <key> in <json_string>, 
    without parsing the json object.  If <key> is not found
    return <default>, or if unspecified raise KeyError.

    Limitations:
        - <json_string> must be a plain json object e.g. '{"key":"val"...}'
        - the target value must be simple, e.g. null, number, or string
        - the first match for <key> (having a simple val) is returned 
            (it is ok if key arises within a nested object)

    Output:
        - all numbers are returned as floats. cast to int as desired.
        - null is returned as None
        - if the key is not found, KeyError is raised, unless <default>
            is specified, in which case <default> is returned
        - keys which have non-simple values (arrays or nested objects) are
            ignored (this can lead to KeyError if key is present but val
            is not simple).
    '''

    find_pattern = (
        r'"%s"'                    # match the key
        r'\s*:\s*'                # match the colon
        r'(null|[\d\.]+|".*?")'    # match the val (null, number, or string)
    ) % key
    find_re = re.compile(find_pattern)

    try:
        val_str = find_re.search(json_string).groups()[0]
    except (AttributeError, IndexError):
        if isinstance(default, SENTINAL):
            raise KeyError(key)
        else:
            return default

    if val_str == 'null':
        return None
    if val_str.startswith('"'):
        return val_str[1:-1]    # strip quotes
    return float(val_str)






import os
import re
import subprocess

def cpus():
	""" Number of available virtual or physical CPUs on this system, i.e.
	user/real as output by time(1) when called with an optimally scaling
	userspace-only program"""

	# cpuset
	# cpuset may restrict the number of *available* processors
	try:
		m = re.search(r'(?m)^Cpus_allowed:\s*(.*)$',
					  open('/proc/self/status').read())
		if m:
			res = bin(int(m.group(1).replace(',', ''), 16)).count('1')
			if res > 0:
				return res
	except IOError:
		pass

	# Python 2.6+
	try:
		import multiprocessing
		return multiprocessing.cpu_count()
	except (ImportError, NotImplementedError):
		pass

	# http://code.google.com/p/psutil/
	try:
		import psutil
		return psutil.cpu_count()   # psutil.NUM_CPUS on old versions
	except (ImportError, AttributeError):
		pass

	# POSIX
	try:
		res = int(os.sysconf('SC_NPROCESSORS_ONLN'))

		if res > 0:
			return res
	except (AttributeError, ValueError):
		pass

	# Windows
	try:
		res = int(os.environ['NUMBER_OF_PROCESSORS'])

		if res > 0:
			return res
	except (KeyError, ValueError):
		pass

	# jython
	try:
		from java.lang import Runtime
		runtime = Runtime.getRuntime()
		res = runtime.availableProcessors()
		if res > 0:
			return res
	except ImportError:
		pass

	# BSD
	try:
		sysctl = subprocess.Popen(['sysctl', '-n', 'hw.ncpu'],
								  stdout=subprocess.PIPE)
		scStdout = sysctl.communicate()[0]
		res = int(scStdout)

		if res > 0:
			return res
	except (OSError, ValueError):
		pass

	# Linux
	try:
		res = open('/proc/cpuinfo').read().count('processor\t:')

		if res > 0:
			return res
	except IOError:
		pass

	# Solaris
	try:
		pseudoDevices = os.listdir('/devices/pseudo/')
		res = 0
		for pd in pseudoDevices:
			if re.match(r'^cpuid@[0-9]+$', pd):
				res += 1

		if res > 0:
			return res
	except OSError:
		pass

	# Other UNIXes (heuristic)
	try:
		try:
			dmesg = open('/var/run/dmesg.boot').read()
		except IOError:
			dmesgProcess = subprocess.Popen(['dmesg'], stdout=subprocess.PIPE)
			dmesg = dmesgProcess.communicate()[0]

		res = 0
		while '\ncpu' + str(res) + ':' in dmesg:
			res += 1

		if res > 0:
			return res
	except OSError:
		pass

	raise Exception('Can not determine number of CPUs on this system')





#!/usr/bin/env python
'''
This manually tests whether PersistentOrderedDict makes a clean exit when 
a script is terminated while it is trying to sync to file.

To test it, run this file with the arugment 'pod', and hit Ctl-C when 
prompted to.  The script will exit with an Exception (but not a 
KeyboardInterrupt).  Then, try running the script again.  If the exit was 
graceful, execution should appear to be the same.  If the exit was not 
graceful, the script will have an exception, with a complaint from 
PersistentOrdederedDict that the file was corrupted

To do an equivalent test for ProgressTracker, follow the same steps as 
above, but use the argument 'tracker' when running this script.

The exit should always be graceful.  If the line within __init__() for
PersistentOrderedDict starting with "atexit" is commented out, this
manual test will no longer pass.
'''


import sys
import t4k


def run_tracker_test():
	tracker = t4k.ProgressTracker('test-dict')

	print 'initializing...'

	tracker.hold()
	for i in range(5000):
		tracker[str(i)] = {'test':False}
	tracker.unhold()

	print 'Done.  Hit Ctl+C !'
	try:
		for i in range(4000):
			tracker[str(i)] = {'test':True}
	except KeyboardInterrupt:
		print 'Good. Now an error should get thrown...'
	else:
		print 'You weren\'t fast enough!'

	tracker = t4k.ProgressTracker('test-dict')

	print 'Hmm... maybe you took to long to hit Ctr+C...'


def run_pod_test():
	pod = t4k.PersistentOrderedDict('test-dict')

	print 'initializing...'

	pod.hold()
	for i in range(5000):
		pod[str(i)] = {'test':False}
	pod.unhold()

	print 'Done.  Hit Ctl+C !'
	try:
		for i in range(4000):
			pod[str(i)] = {'test':True}
	except KeyboardInterrupt:
		print 'Good. Now an error should get thrown...'
	else:
		print 'You weren\'t fast enough!'

	pod = t4k.PersistentOrderedDict('test-dict')

	print 'Hmm... maybe you took to long to hit Ctr+C...'


if __name__ == '__main__':
	usage = (
		'\nusage:\n\t./test_clean_exit.py pod'
		'\n\t./test_clean_exit.py tracker\n'
	)

	try:
		if sys.argv[1] == 'pod':
			run_pod_test()
		elif sys.argv[1] == 'tracker':
			run_tracker_test()
		else:
			print usage

	except IndexError:
		print usage



import multiprocessing
from multiprocessing.managers import BaseManager


class MyManager(BaseManager):
	pass

class MyObject(object):
	def __init__(self):
		self.val = 0

	def inc(self):
		self.val += 1

	def get(self):
		return self.val


MyManager.register('MyObject', MyObject, exposed=['get', 'inc'])


def test_manager():
	manager = MyManager()
	manager.start()
	my_object = manager.MyObject()

	p1 = multiprocessing.Process(target=do_increment, args=(my_object,'A'))
	p2 = multiprocessing.Process(target=do_increment, args=(my_object,'B'))

	p1.start()
	p2.start()

	p1.join()
	p2.join()

	print my_object.get()


def do_increment(obj, name):
	for i in range(10000):

		if i % 1000 == 0:
			print name

		obj.inc()


if __name__=='__main__':
	test_manager()



def _filter_min_max_args(*args):

	if len(args) == 1:
		args = args[0]

	# filter out arguments that are None
	try:
		filtered_args = [i for i in args if i is not None]

	# arguments not iterable (i.e. a single non-iterable value was passed)
	except TypeError:
		if args is None:
			return None
		else:
			filtered_args = args

	else:
		# if argument was iterable, but empty after filtering, return None
		if len(filtered_args) == 0:
			return None

		# if argument was iterable, but is singleton now, wrap as list
		elif len(filtered_args) == 1:
			return [filtered_args]

	return filtered_args


def safe_max(*args):

	# filter out None's
	args = _filter_min_max_args(*args)
	if args is None:
		return None

	# delegate to max
	return max(*args)


def safe_min(*args):

	# filter out None's
	args = _filter_min_max_args(*args)
	if args is None:
		return None

	# delegate to min
	return min(*args)


def safe_lte(val1, val2):
	if val1 is None:
		return False
	if val2 is None:
		return True
	return val1 <= val2

def safe_lt(val1, val2):
	if val1 is None:
		return False
	if val2 is None:
		return True
	return val1 < val2

def safe_gte(val1, val2):
	if val1 is None:
		return False
	if val2 is None:
		return True
	return val1 >= val2

def safe_gt(val1, val2):
	if val1 is None:
		return False
	if val2 is None:
		return True
	return val1 > val2



# Builtins
import os
import json
from collections import defaultdict
from collections import Counter


class CrowdflowerResults(object):
	"""
	Reads in and represents crowdflower results stored in json files located at 
	``paths``.  ``paths`` can be a single json filename or multiple filenames.
	This class is list-like (it wraps a list of rows found in the original
	results sets, and exposes the methods of that list).

	When multiple filenames are given, the results from these files are merged.
	Rows in these results that yield the same value when passed into
	``merge_key`` are considered to hold data about the same question, and will
	be merged into a single row in the final representation.  

	By default gold rows are excluded.  use ``exclude_gold=False`` to include
	gold rows.
	
	Access rows by indexing into the CrowdflowerResults object as if it were a
	list.  Each item is a row, similar to what appears in the original results
	file, but augmented with a few additional keys:

		``count``: contains a counter that indicates how many times each
			response was given for every field in the question.

		``fraction``: contains a dictionary that indicates what proportion of
			times each response was given for every field in the question.

		``mode``: contains a dictionary that indicates the most common response
			for each field in the question.  If multiple responses are tied,
			then they are all listed.  For consistency, when one response is
			most common, it is still presented as a length-1 list.

	You can also access rows based on a dictionary structure, using the id they 
	return when passed into ``merge_key``, by using the CrowdflowerResult.dict
	attribute.
	"""

	def __init__(
		self,
		paths,
		merge_key=None,
		exclude_gold=True
	):
		results = read_results(paths, merge_key)
		self.augmented_results = augment_results(results, exclude_gold)

		if merge_key is not None:
			self.dict = { merge_key(row):row for row in self.augmented_results}

	def __getattr__(self, attr):
		return getattr(self.augmented_results, attr)

	def __getitem__(self, key):
		return self.augmented_results.__getitem__(key)

	def __iter__(self):
		return iter(self.augmented_results)

	def __len__(self):
		return len(self.augmented_results)



def read_augmented_results(
	paths, exclude_gold=True, question_id_key='question_id'
):
	"""
	Read in crowdflower results stored in json files at locations specified by
	paths.  paths may either be a single path or iterable of paths, in which
	case results for multiple tasks will be merged.  

	Merging results is accomplished by considering two rows from different
	results files as corresponding to the same question if they have the same
	id, which is a value located in the data row defining the question in the
	column designated by "question_id_key".

	After merging results, augment them, by computing the number of votes for
	each answer in each question, computing the fraction of votes for each
	answer, and computing the most common response.
	"""
	if isinstance(paths, basestring):
		paths = [paths]

	return augment_results(read_results(paths), exclude_gold)


def merge_results(result_sets, merge_key):
	"""
	Merges the result_sets listed in ``result_sets``.  Two rows are considered
	to correspond to the same question if they return the same value when
	passed into ``merge_key``.
	"""

	# If there are no results, return an empty list
	if len(result_sets)<1:
		return []

	# If there are less than two results, then there's no merging to do, just
	# return the single result set
	if len(result_sets) < 2:
		return result_sets[0]

	# We'll use a dictionary structure to merge results having the same
	# question id.  After merging, we'll convert back to a list.  To merge
	# results, all we do is append judgments.  Note that this invalidates the
	# aggregate statistics that the crowdflower platform inserts into results,
	# which are therefore removed, and aren't very useful anyway.
	merged = {}
	invalid_fields = [
		'agreement', 'created_at', 'updated_at', 'state', 
		'gold_pool', 
	]

	for result_set in result_sets:
		for result in result_set:
			question_id = merge_key(result)

			# If this question id is already in the merged results, all we want
			# to do is add the judgments from this result for that question
			if question_id in merged:

				merged_q = merged[question_id]

				# Add this result's judgments to the merged judgments for this
				# question
				merged_judgments = merged_q['results']['judgments']
				add_judgments = result['results']['judgments']
				merged_judgments.extend(add_judgments)

				# Increment counters associated to this question accordingly
				merged_q['judgments_count'] += result['judgments_count']
				merged_q['missed_count'] += result['missed_count']

				# Normally each row has a crowd-flower assigned id, and it also
				# records the job id it's from.  We'll aggregate the
				# CF-assigned ids and the ids of the jobs in which this
				# question appeared
				merged_q['id'].append(result['id'])
				merged_q['job_id'].append(result['job_id'])


			# If this question is not already in the merged results, then we'll
			# add it.  We need to remove certain fields that won't be valid
			# after merging has taken place.
			else:

				# Delete the invalid fields
				for invalid_field in invalid_fields:
					del result[invalid_field]

				# Delete aggregated responses, keep only individual judgments
				for aggregate_result_field in result['results'].keys():
					if aggregate_result_field != 'judgments':
						del result['results'][aggregate_result_field]

				# Convert certain fields to lists, so the values from different
				# result-sets can be accumulated
				result['id'] = [result['id']]
				result['job_id'] = [result['job_id']]

				# Add to the merged results
				merged[question_id] = result

	# Now that we've merged the results, we don't need them in dictionary form.
	# Restore to list form for consistency with non-merged result sets.
	return merged.values()



def read_results(paths, merge_key=None):
	"""
	Read crowdflower results json file(s) located at paths.

	paths can be a single path or iterable of paths.  If an iterable of paths
	is supplied, then the results from multiple files will be merged.  
	
	Rows from different files are considered to be about the same question
	if the same value is returned when ``merge_key`` is called on those rows.

	If only a single path is given, then no merging is actually done, and
	``merge_key`` will never be called.
	"""

	# Handle the fact that a single path may be provided, rather than an
	# iterable of paths.
	if isinstance(paths, basestring):
		paths = [paths]

	# Read results form each file and merge them
	result_sets = []
	for path in paths:
		result_sets.append([json.loads(l) for l in open(path)])

	# If a method for assigning id's to rows is provided, then rows having the
	# same id will be merged, accumulating judgments for both
	if merge_key is not None:
		return merge_results(result_sets, merge_key)

	# Otherwise, all the rows from all results sets are returned without any
	# attempt to merge rows that pertain to the same question.
	return sum(result_sets, [])



def augment_results(results, exclude_gold = True):
	"""
	Augments CrowdFlower results by adding a few more keys to each row 
	(i.e. results to each question), which
	help summarize the results obtained.

	Each row corresponds to a single question, normally shown to multiple
	annotators.  Each question has multiple "fields" corresponding to each
	of the different CML inputs that annotators fill out for a given question.

	Because multiple annotators answer a single question, we get multiple
	answers for each field in each question.  This summarises the results, by
	providing counts for the number of times each answer is seen, the fraction
	of times each answer is seen, and the most common answer(s).

	These are directly added to the results object, thus they cause a side
	effect.  The modified results object is also returned.  The following keys
	are added to each row of the results object:

		row['counts']
			A dictionary whose keys are the names of the fields in the
			questions, and whose values are counters indicating how many times
			each response value is seen for the field.

		row['fractions']
			A dictionary whose keys are the names of the fields, and whose
			values are dictionaries that indicate the fraction of times
			different responses were seen for that field.

		row['mode']
			A dictionary whose keys are the names of the fields, and whose
			values are a list of the most common response(s) for that field.
			More than one value appears in the list when more than one response
			is tied as "most common".

	Additionally, when exclude gold is True, the gold rows are deleted from the
	results.
	"""

	# For each question, record how many votes were cast for each quote
	# as being the most / least verifiable.  Also record the attribution
	# ids for the quotes.
	augmented_results = []
	for result in results:

		# Ignore test questions
		if '_golden' in result['data'] and exclude_gold:
			continue

		# Accumulate information in this reportable result entry
		counts = defaultdict(Counter)
		mode = {}
		fraction = defaultdict(dict)

		# Count each type of response for each field
		for judgment in result['results']['judgments']:
			for field, val in judgment['data'].items():
				counts[field][val] += 1

		# Having accumulated the counts for each value for each field, use that
		# to calculate the mode value and the fraction of times each value is
		# given for each field
		for field in counts:

			# Figure out the mode value for this field.  Actually we
			# return a list of mode values, because more than one
			# value can tie as the most common value
			pointer = 0
			mode[field] = []
			sorted_counts = counts[field].most_common()
			while (
				pointer < len(sorted_counts) 
				and sorted_counts[pointer][1] == sorted_counts[0][1]
			):
				mode[field].append(sorted_counts[pointer][0])
				pointer += 1

			# Store the fraction of times a given answer is given for a field.
			# I.e. this is the same as counts, but is normalized to
			# sum to 1.
			total = float(sum(counts[field].values()))
			for val in counts[field]:
				fraction[field][val] = counts[field][val]/total

		# Put the counts, mode, and fraction onto the autmented result, and store
		# add that augmented result to the autmented results set.
		augmented_result = dict(result)
		augmented_result.update({
			'counts':counts, 'mode':mode, 'fraction':fraction})
		augmented_results.append(augmented_result)

	return augmented_results




from safe import safe_min, safe_lte
from grouper import trim


def string_distance(s1, s2):
	return SA.string_distance(s1, s2)

def string_align(s1, s2):
	distance, path = SA.string_alignment(s1, s2)
	return distance, path

def string_align_masks(s1, s2):
	mask1, mask2 = SA.get_string_alignment_masks(s1, s2)
	return 

def string_align_path(s1, s2):
	return SA.get_string_alignment_path(s1, s2)

def alignment_score(s1, s2, as_substring=False):
	return SA.alignment_score(s1, s2, as_substring)


class StringAligner(object):

	UP = 1
	LEFT = 2
	DIAG = 3

	DELETE = 'd'
	INSERT = 'i'
	MATCH = 'm'
	SUBSTITUTE = 's'

	def __init__(
		self,
		case_sensitive=True,
		deletion_penalty=1,
		insertion_penalty=1,
	):
		self.deletion_penalty = deletion_penalty
		self.insertion_penalty = insertion_penalty
		self.case_sensitive = case_sensitive


	def string_alignment(self, name1, name2, substring_mode=False):
		distance = [
			[None for j in range(len(name2)+1)] for i in range(len(name1)+1)
		]
		path = [
			[None for j in range(len(name2)+1)] for i in range(len(name1)+1)
		]

		# Normalize case if option chosen for case insensitive matches
		if not self.case_sensitive:
			name1 = name1.lower()
			name2 = name2.lower()

		# Dynamic program computes the score for all possible sequences
		# of matches, deletions, insertions, and substitutions that
		# align the strings
		for i in range(len(name1) + 1):
			for j in range(len(name2) + 1):

				# The word boundary matches
				if i == 0 and j == 0:
					distance[i][j] = 0
					path[i][j] = self.DIAG
					continue

				# match character i and j
				match_dist = None
				if i > 0 and j > 0:
					if name1[i-1] == name2[j-1]:
						match_dist = distance[i-1][j-1]

					# substitute character i by char j
					else:
						match_dist = (
							distance[i-1][j-1] 
							+ self.deletion_penalty 
							+ self.insertion_penalty
						)

				# Delete one char from name1
				del_dist = None
				if i > 0:

					# In substring mode, leading and trailing deletions
					# from name1 are penalty-free
					if substring_mode and (j == 0 or j == len(name2)):
						del_dist = distance[i-1][j]

					# Otherwise the deletion penalty of 1 is applied
					else:
						del_dist = distance[i-1][j] + self.deletion_penalty

				# Insert one char from name2
				insert_dist = None
				if j > 0:
					insert_dist = distance[i][j-1] + self.insertion_penalty

				distance[i][j] = safe_min(
					match_dist, del_dist, insert_dist)

				if safe_lte(match_dist, del_dist):
					if safe_lte(match_dist, insert_dist):
						path[i][j] = self.DIAG
					else:
						path[i][j] = self.LEFT

				elif safe_lte(del_dist, insert_dist):
					path[i][j] = self.UP

				else:
					path[i][j] = self.LEFT

		return distance, path


	def alignment_score(self, reference_string, query_string, as_substring):
		alignment_path = self.get_string_alignment_path(
			reference_string, query_string, as_substring)
		
		if as_substring:
			alignment_path = trim(alignment_path, 'd')

		substitutions = sum([x == 's' for x in alignment_path])
		deletions = sum([x == 'd' for x in alignment_path])
		insertions = sum([x == 'i' for x in alignment_path])
		matches = sum([x == 'm' for x in alignment_path])

		score = (
			matches 
			- insertions * self.insertion_penalty
			- deletions * self.deletion_penalty
			- substitutions * (
				self.deletion_penalty + self.insertion_penalty)
		)

		return score


	def string_distance(self, name1, name2, substring_mode=False):
		distance, path = self.string_alignment(name1, name2, substring_mode)
		return distance[-1][-1]



	def get_string_alignment_masks(
		self,
		name1,
		name2,
		substring_mode=False
	):
		distance, path = self.string_alignment(name1, name2, substring_mode)

		alignment1 = [None for p in range(len(name1))]
		alignment2 = [None for q in range(len(name2))]
		i = len(name1)
		j = len(name2)
		while i > 0 or j > 0:
			if path[i][j] == self.UP:
				i -= 1
				alignment1[i] = 0
				continue

			if path[i][j] == self.LEFT:
				j -= 1
				alignment2[j] = 0
				continue

			if path[i][j] == self.DIAG:
				i -= 1
				j -= 1
				if distance[i][j] == distance[i+1][j+1]:
					alignment1[i] = 1
					alignment2[j] = 1
				else:
					alignment1[i] = 0
					alignment2[j] = 0

		return alignment1, alignment2


	def get_string_alignment_path(self, name1, name2, substring_mode=False):
		distance, path = self.string_alignment(name1, name2, substring_mode)

		linear_path = []
		i = len(name1)
		j = len(name2)
		while i > 0 or j > 0:
			if path[i][j] == self.UP:
				i -= 1
				linear_path.append(self.DELETE)
				continue

			if path[i][j] == self.LEFT:
				j -= 1
				linear_path.append(self.INSERT)
				continue

			if path[i][j] == self.DIAG:
				i -= 1
				j -= 1
				if distance[i][j] == distance[i+1][j+1]:
					linear_path.append(self.MATCH)
				else:
					linear_path.append(self.SUBSTITUTE)

		linear_path.reverse()
		return linear_path


	def display_string_alignment(self, name1, name2, substring_mode=False):
		alignment1, alignment2 = self.get_string_alignment_masks(
			name1, name2, substring_mode
		)

		i = 0
		j = 0
		display1 = []
		display2 = []
		display = []
		while i < len(name1) or j < len(name2):

			if i >= len(name1):
				display1.append('-')
				display2.append(name2[j])
				display.append(' ')
				j += 1
				continue

			if j >= len(name2):
				display1.append(name1[i])
				display2.append('-')
				display.append(' ')
				i += 1
				continue

			if alignment1[i]:
				if alignment2[j]:
					display1.append(name1[i])
					display2.append(name2[j])
					display.append('|')
					i += 1
					j += 1

				else:
					display1.append('-')
					display2.append(name2[j])
					display.append(' ')
					j += 1

			else:
				if alignment2[j]:
					display1.append(name1[i])
					display2.append('-')
					display.append(' ')
					i += 1

				else:

					# In substring mode, we favor deletion from the 
					# reference string (name1) until at least one match
					# is seen
					if substring_mode and j==0:
						display1.append(name1[i])
						display2.append('-')
						display.append(' ')
						i += 1

					# But normally this would be interpreted as a 
					# substitution
					else:
						display1.append(name1[i])
						display2.append(name2[j])
						display.append(' ')
						i += 1
						j += 1

		return '\n'.join([
			''.join(display1),
			''.join(display),
			''.join(display2)
		])

SA = StringAligner()



import t4k

def invert_dict(dictionary):
    new_dict = {}
    for key in dictionary:
        new_dict[dictionary[key]] = key

    return new_dict



def max_item(dictionary):
    """
    This conceptually like an argmax function -- it finds the key whose
    corresponding value is the largest in the dictionary.
    """
    the_max = t4k.Max()
    for key, value in dictionary.iteritems(): the_max.add(value, key)
    large_val, large_key = the_max.get()
    return large_key, large_val


def min_item(dictionary):
    """
    This conceptually like an argmax function -- it finds the key whose
    corresponding value is the largest in the dictionary.
    """
    the_min = t4k.Min()
    for key, value in dictionary.iteritems(): the_min.add(value, key)
    large_val, large_key = the_min.get()
    return large_key, large_val


def max_key(dictionary):
    """
    This conceptually like an argmax function -- it finds the key whose
    corresponding value is the largest in the dictionary.
    """
    return max_item(dictionary)[0]


def min_key(dictionary):
    """
    This conceptually like an argmin function -- it finds the key whose
    corresponding value is the smallest in the dictionary.
    """
    return min_item(dictionary)[0]


def max_value(dictionary):
    """
    Returns the maximum value in the dictionary.  Equivalent to
    max(dictionary.values()).
    """
    return max_item(dictionary)[1]


def min_value(dictionary):
    """
    Returns the minimum value in the dictionary.  Equivalent to
    min(dictionary.values()).
    """
    return min_item(dictionary)[1]



def merge_dicts(*dictionaries, **kwargs):
    """
    Merge an arbitrary number of dictionaries, in such a way that values found
    in dictionaries that appear earlier in the parameter list take precedence.

    If the only kwarg ``recursive`` is True, then sub-dictionaries are merged
    in the same way, otherwise a dictionary-valued entries completely overwrite
    one another.
    """
    recursive = kwargs.pop('recursive', True)
    if recursive:
        return merge_dicts_recursive(*dictionaries)
    return merge_dicts_nonrecursive(*dictionaries)


def merge_dicts_nonrecursive(*dictionaries):
    """
    Merge an arbitrary number of dictionaries, in such a way that values found
    in dictionaries that appear earlier in the parameter list take precedence.
    Note that sub-dictionaries are treated as simple values -- the
    sub-dictionary that takes precedence completely overwrites the other(s),
    rather than being recursively merged.
    """
    merged = {}
    for dictionary in reversed(dictionaries):
        merged.update(dictionary)
    return merged


def merge_dicts_recursive(*dictionaries):
    """
    Merge an arbitrary number of dictionaries, in such a way that values found
    in dictionaries that appear earlier in the parameter list take precedence.
    Note that sub-dictionaries are recursively merged in the same way.
    """
    merged = {}
    for dictionary in reversed(dictionaries):
        for key in dictionary:

            # If this key corresponds to a sub-dictionary, both within the
            # merged dict so far and in the current dict to merge in, then we
            # recurse, so that those sub-dicts are merged well 
            both_vals_are_dicts = (
                key in merged and isinstance(merged[key], dict) 
                and isinstance(dictionary[key], dict)
            )

            if both_vals_are_dicts:
                merged[key] = merge_dicts(dictionary[key], merged[key])

            else:
                merged[key] = dictionary[key]

    return merged


def dzip(*dictionaries):
    '''
    Like zip, but for dictionaries.  Produce a dictionary whose keys are 
    given by the intersection of input dictionaries' keys, and whose
    values are are tuples of the input dicts corresponding values.
    '''

    # Define the dzip of no dictionaries to be an empty dictionary
    if len(dictionaries) == 0:
        return {}

    # Get the keys common to all dictionaries
    keys = set(dictionaries[1])
    for d in dictionaries[1:]:
            keys &= set(d)

    # Make the zipped dictionary
    return {
        key : tuple([d.get(key, None) for d in dictionaries])
        for key in keys
    }


def select(dictionary, fields, require=True):
    """
    Create a new dict by copying selected `fields` from the original
    dict `dictionary`.  If `require` is True, then a field that doesn't exist
    in the original dict will cause a KeyError; otherwise it passes over
    missing keys
    silently.
    """
    new_dict = {}
    for field in fields:
        try:
            new_dict[field] = dictionary[field]
        except KeyError:
            if require:
                raise

    return new_dict





class Max(object):

	def __init__(self, keep_last=True):
            self.extreme_key = None
            self.extreme_item = None
            self.keep_last = keep_last
            self._is_draw = False

	def add(self, key, item):

            # The first item to ever be checked always wins
            if self.extreme_key is None:
                is_greater = True
                is_lesser = False

            # Later items get compared to the current winner.
            else:
                is_greater = self.comp(key, self.extreme_key)
                is_lesser = self.comp(self.extreme_key, key)

            # If we have a new winner, remember it.  It also means we don't
            # have a draw
            if is_greater:
                self._is_draw = False
                self.extreme_key = key
                self.extreme_item = item

            # If we don't have a winner, but it's also not a loser, then its a
            # tie.  Depending on the setting of ``keep_last``, we might keep
            # the newer item instead of the prior one.
            elif not is_lesser:
                self._is_draw = True
                if self.keep_last:
                    self.extreme_key = key
                    self.extreme_item = item


        def is_draw(self):
            return self._is_draw


	def update(self, key_item_pairs):
		for key, item in key_item_pairs:
			self.add(key, item)

	def get(self):
		return self.extreme_key, self.extreme_item

	def comp(self, key1, key2):
		return key1 > key2


class Min(Max):
	def comp(self, key1, key2):
		return key1 < key2



#!/usr/bin/env python
from selenium_crawler import SeleniumCrawler, uses_selenium


class TestCrawler(SeleniumCrawler):

	HOMEPAGE = 1
	ERROR_PAGE = 2

	@uses_selenium
	def crawl(self):

		# For a demo, let's get a couple web pages 
		urls = ['http://ubuntu.com', 'http://ubuntu.com/fake']
		for url in urls:

			# Use delay() to avoid making requests too often
			self.delay()

			# Get a web page
			print '\nfetching %s' % url
			self.driver.get(url)

			# Recognize the kind of page 
			page_type = self.wait(self.recognize_page)
			if page_type == self.HOMEPAGE:
				print 'got the home page'
			elif page_type == self.ERROR_PAGE:
				print 'got a 404 page'

			# Get the page source
			print "here's a snippet of page source:"
			print self.get_page_source()[:100]

		print '\nDone!'


	def recognize_page(self):

		# Is it a profile page?
		try:
			self.driver.find_element_by_class_name('homepage')
			return self.HOMEPAGE
		except self.NoSuchElementException:
			pass

		# Is it a 404 page?
		try:
			self.driver.find_element_by_class_name('error404')
			return self.ERROR_PAGE
		except self.NoSuchElementException:
			pass

		return False

if __name__ == '__main__':
	crawler = TestCrawler()
	crawler.crawl()



#!/usr/bin/env python

import psutil
import time
from multiprocessing import Process, Pipe

DONE = True
NOT_DONE = False
SUICIDE = -1
DEFAULT_KILL_TIMEOUT = 5

class ManagedProcess(object):
    '''
    Runs a target function in a subprocess and monitors it.
    The subprocess must call back through a pipe periodically.  If the
    manager doesn't hear from the subprocess after a specified time delay,
    it assumes the subprocess is hung, and so kills the process, and 
    then restarts it by calling the target function again.  

    When the target function completes successfully, it should call back
    to the manager and signal that it is done.
    '''

    def __init__(
        self,
        target,
        args=(),
        kwargs={},
        timeout=60,
        kill_timeout=DEFAULT_KILL_TIMEOUT
    ):

        # Register args
        self.target = target
        self.args = args
        self.kwargs = kwargs
        self.timeout = timeout


    def _start_managed_process(self):

        # Make a pipe for communication with the subprocess
        self.connection, child_connection = Pipe()
        self.kwargs['pipe'] = child_connection

        # Call target as a separate process
        self.managed_process = Process(
            target=self.target,
            args=self.args,
            kwargs=self.kwargs
        )
        self.managed_process.start()


    def start(self):
        manager_process = Process(target=self._start)
        manager_process.start()


    def _start(self):
        '''
        call <func> inside a subprocess, passing it <args> and <kwargs>.  
        Expect that subprocess to communicate via a pipe every 
        <timeout> seconds.  If no call back is received after 
        <timeout> seconds, kill the subprocess, and restart it
        by calling func again with the original arguments.

        The target function must have an argument called pipe, which
        will be passed one end of a multiprocessing.Pipe().  It doesn't
        matter where that argument occurs in the function signature,
        but any arguments after 'pipe' should be filled using <kwargs>.
        '''

        # Start the subprocess
        self._start_managed_process()

        # Monitor it.  Restart if it crashes or hangs.
        status = NOT_DONE
        while status == NOT_DONE:

            # Listen for a call from the subprocess
            responding = self.connection.poll(self.timeout)

            # Update status if we get a response
            if responding:
                response = self.connection.recv()
                if response == DONE:
                    status = DONE

            # If no response, or suicide, restart the process
            if not responding or response == SUICIDE:

                # Give suicidal processes time to die
                if response == SUICIDE:
                    time.sleep(3)

                # Try to forcibly kill the process
                try:
                    kill_proc_tree(
                                            self.managed_process.pid,
                                            kill_timeout=self.kill_timeout
                                        )

                # If the process was already dead, that's OK
                except psutil.NoSuchProcess:
                    pass

                # Restart the process
                print 'Manager: respawning process...'
                self._start_managed_process()

        print 'Manager: process finished!'


def kill_proc_tree(pid, kill_timeout=DEFAULT_KILL_TIMEOUT):    
    '''
    Kill the process having id <pid> and kill all of it's children.
    '''

    # get the process
    parent = psutil.Process(pid)

    # get the children recursively
    children = parent.children(recursive=True)

    # kill all children
    for child in children:
        child.kill()
    psutil.wait_procs(children, timeout=kill_timeout)

    # now kill the parent
    parent.kill()
    parent.wait(kill_timeout)






import csv, codecs, cStringIO


class UTF8Recoder:
    """
    Iterator that reads an encoded stream and reencodes the input to UTF-8
    """
    def __init__(self, f, encoding):
        self.reader = codecs.getreader(encoding)(f)

    def __iter__(self):
        return self

    def next(self):
        return self.reader.next().encode("utf-8")


class UnicodeTsvReader:
    """
    A CSV reader which will iterate over lines in the CSV file "f",
    which is encoded in the given encoding.
    """

    def __init__(
		self, 
		f, 
		dialect=csv.excel,
		encoding="utf-8",
		delimiter='\t',
		**kwds
	):
        f = UTF8Recoder(f, encoding)
        self.reader = csv.reader(
			f, dialect=dialect, delimiter=delimiter, **kwds
		)

    def next(self):
        row = self.reader.next()
        return [unicode(s, "utf-8") for s in row]

    def __iter__(self):
        return self


class UnicodeTsvWriter:
    """
    A CSV writer which will write rows to CSV file "f",
    which is encoded in the given encoding.
    """

    def __init__(
		self,
		f,
		dialect=csv.excel,
		encoding="utf-8",
		delimiter='\t',
		**kwds
	):
        # Redirect output to a queue
        self.queue = cStringIO.StringIO()
        self.writer = csv.writer(
			self.queue, dialect=dialect, delimiter=delimiter, **kwds
		)
        self.stream = f
        self.encoder = codecs.getincrementalencoder(encoding)()

    def writerow(self, row):
        self.writer.writerow([s.encode("utf-8") for s in row])
        # Fetch UTF-8 output from the queue ...
        data = self.queue.getvalue()
        data = data.decode("utf-8")
        # ... and reencode it into the target encoding
        data = self.encoder.encode(data)
        # write to the target stream
        self.stream.write(data)
        # empty queue
        self.queue.truncate(0)

    def writerows(self, rows):
        for row in rows:
            self.writerow(row)



def pc(fraction, decimals=2):
    formatter = '%%.%df %%%%' % decimals
    return formatter % (fraction*100)

def progress(i, total, period=1000, prefix='', decimals=1):
	'''
	Prints a string representing the % completion when starting the
	`i`th step out of `total` steps, but only prints every `period` steps.
	'''
        formatter = '%s%2.' + str(decimals) + 'f %%'
	if i % period == 0:
		print formatter % (prefix, i/float(total) * 100)



from datetime import date, datetime, timedelta
class DateIterator(object):
	'''
	This iterable makes it easy to walk between two dates, based on an
	interval specified in days.  In each iteration, the iterator yields
	a tuple giving the starting and ending date for that interval.

	Dates are input as Datetime.date or datetime.datetime objects, and 
	are output as datetime.date objects.

	To iterate over every day between two dates, provide the first and
	last day, and set the increment_by_days keyword argument to 1
	'''

	CHRONOLOGICAL = 1
	STATIONARY = 0
	REVERSE_CHRONOLOGICAL = -1

	def __init__(self, start, end, increment_by_days=1):

		# Register variables
		self.start = start
		self.end = end
		self.increment = increment_by_days 

		# ensure that start and end are date-like
		are_start_end_date_like = all([
			hasattr(start, 'year'),
			hasattr(start, 'month'),
			hasattr(start, 'day'),
			hasattr(end, 'year'),
			hasattr(end, 'month'),
			hasattr(end, 'day')
		])

		if not are_start_end_date_like:
			raise ValueError(
				'<start> and <end> should be like datetime.date '
				'or datetime.datetime.'
			)

		# what direction is this iterator moving in?
		if self.increment > 0:
			self.direction = self.CHRONOLOGICAL
		elif self.increment < 0:
			self.direction = self.REVERSE_CHRONOLOGICAL
		elif self.increment == 0:
			self.direction = self.STATIONARY

		# this is the difference between the first day and last day
		# within one date increment.  It is correctly 0 if increment
		# is either -1, 0, or 1.  
		# Generally 1 less than increment in magnitude
		self.span = self.direction * (abs(increment_by_days) - 1)

		# make sure that increment will move from start towards end
		if self.start > self.end and self.direction >= 0:
			raise ValueError(
				'If start date is after end date, the '
				'increment parameter should be positive.'
			)
		elif self.start < self.end and self.direction <=0:
			raise ValueError(
				'If start date is before end date, the '
				'increment parameter should be negative.'
			)

		self.pointer = start


	def __iter__(self):

		# the iterator is a fresh copy of this.
		# this means that one can instantiate multiple iterators from the
		# same DateIterator instance concurrently
		return self.__class__(self.start, self.end, self.increment)


	def next(self):

		# If we are alread passed the end date, stop iterating
		if self.pointer > self.end and self.direction >= 0:
			raise StopIteration
		elif self.pointer < self.end and self.direction <= 0:
			raise StopIteration

		# Determine the starting and ending dates for this subinterval
		interval_start = self.pointer
		interval_end = self.pointer + timedelta(self.span)

		# If interval end falls outside the range, truncate it
		if interval_end > self.end and self.direction >= 0:
			interval_end = self.end
		elif interval_end < self.end and self.direction <= 0:
			interval_end = self.end

		# Update the internal pointer
		self.pointer += timedelta(self.increment)

		# Return the subinterval
		return interval_start, interval_end


class DateBinner(object):

	DATE_FORMAT = '%Y/%m/%d'

	def __init__(self, start, end, increment_by_days=7):
		self.start = start
		self.end = end
		self.increment_by_days = increment_by_days

		self.date_keys = self.get_date_keys(start, end, increment_by_days)


	def get_date_keys(self, start, end, increment_by_days):

		# We'll accumulate a list of tuples consisting of <k, start, end>
		# where k is a string (a key) that represents the time interval
		# from start to end (which are dates, inclusive).
		# This will help map an arbitrary date onto the string representing
		# the interval in which that date falls
		date_keys = []

		# Make a date iterator to yield the desired interval endpoints
		date_iterator = DateIterator(start, end, increment_by_days)

		# Use the iterator and accumulate all of the key-interval 
		# assignments
		for interval_start, interval_end in date_iterator:
			start_str = interval_start.strftime(self.DATE_FORMAT)
			end_str = interval_end.strftime(self.DATE_FORMAT)
			key = '%s-%s' % (start_str, end_str)
			date_keys.append((key, interval_start, interval_end))

		return date_keys


	def get_date_key(self, datelike):
		''' 
		Returns a string, representing date range into which <datelike> 
		falls.  <datelike> should be a date or datetime object.  
		'''
		if isinstance(datelike, datetime):
			datelike = datelike.date()
		for key, start, end in self.date_keys:
			if datelike >= start and datelike <= end:
				return key 
		return None



# saved as greeting-server.py
import setproctitle
import sys
import os
import subprocess
import Pyro4
import multiprocessing

Pyro4.config.SERIALIZERS_ACCEPTED = ('pickle',)
Pyro4.config.SERIALIZER = 'pickle'

@Pyro4.expose
class Provider(object):
    def __init__(self, obj_to_provide):
        self.obj_to_provide = obj_to_provide

    def provide(self):
        return self.obj_to_provide


def mem_get(name):
    return Pyro4.Proxy('PYRONAME:%s' % name).provide()


def remember(name, build_function):

    # First, try to get the object from a memory server
    try:
        print 'trying to remember'
        obj = mem_get(name)
        print 'remembered'

    # If that failed, create the object. Put in a memory server for next time
    except (Pyro4.errors.NamingError, Pyro4.errors.CommunicationError):

        # Make the object
        print 'could not remember... creating'
        obj = build_function()

        # Set the process title for the memory server so that it can be
        # managed using the mem-server command line tool
        setproctitle.setproctitle('memory-server-%s' % name)
        def memory_func():
            server = MemoryServer()
            server.provide(name, obj)
            server.start()

        # Start the memory server process as a detached process
        spawn_daemon(memory_func)


    # Return the object
    return obj


def set_proc_name(newname):
    """
    Modify the title of the process displayed when using the ps command.
    This makes it possible to manage memory server processes using the 
    mem-server command line tool, which finds memory server processes based
    on their name
    """
    from ctypes import cdll, byref, create_string_buffer
    libc = cdll.LoadLibrary('libc.so.6')
    buff = create_string_buffer(len(newname)+1)
    buff.value = newname
    libc.prctl(15, byref(buff), 0, 0, 0)


def spawn_daemon(func):
    """
    Start a process that is detached from the parent, and which continues
    after the parent exists.  That process executes the function ``func``,
    which must take no arguments.
    """
    # do the UNIX double-fork magic, see Stevens' "Advanced 
    # Programming in the UNIX Environment" for details (ISBN 0201563177)
    try: 
        pid = os.fork() 
        if pid > 0:
            # parent process, return and keep running
            return
    except OSError, e:
        print >>sys.stderr, "fork #1 failed: %d (%s)" % (e.errno, e.strerror) 
        sys.exit(1)

    os.setsid()

    # do second fork
    try: 
        pid = os.fork() 
        if pid > 0:
            # exit from second parent
            sys.exit(0) 
    except OSError, e: 
        print >>sys.stderr, "fork #2 failed: %d (%s)" % (e.errno, e.strerror) 
        sys.exit(1)

    # do stuff
    func()

    # all done
    os._exit(os.EX_OK)


class MemoryServer(object):
    """
    The MemoryServer class can be used to make python objects available to 
    other processes.  First, instantiate a MemoryServer object, then call
    memory_server.provide(name, obj), giving a name by which clients will find
    the object, followed by the object itself.  After calling provide for any 
    number of objects, call memory_server.start() to make those objects 
    available.

    clients should use mem_get(name) function in this module to retrieve the
    objects, providing the same name given to the MemoryServer instance.
    """

    def __init__(self):
        self.daemon = Pyro4.Daemon() 
        self.name_server = None


    def provide(self, name, obj):
        """
        Exposes the object ``obj`` which can be retrieved using ``name``
        """
        provider = Provider(obj)
        uri = self.daemon.register(provider)
        self._get_name_server().register(name, uri)


    def _get_name_server(self):
        """
        Returns a name server.  If necessary it starts one up.
        """
        # If we already know the name server, return it
        if self.name_server is not None:
            return self.name_server

        # Otherwise, try to find an existing name server
        try:
            self.name_server = Pyro4.locateNS()

        # If there's no existing name server, then start one
        except Pyro4.errors.NamingError:
            self._start_name_server()
            self.name_server = Pyro4.locateNS()

        # Return the name server.
        return self.name_server


    def _start_name_server(self):
        subprocess.Popen(
            ['pyro4-ns', '-n', '0.0.0.0'],
            env=dict(os.environ, PYRO_SERIALIZERS_ACCEPTED='pickle'),
            stdout=open(os.devnull, 'w'),
            close_fds=True
        )


    def start(self):
        self.daemon.requestLoop()






'''
This module keeps track of the vocabulary present in a corpus, provides
a two-way maping between tokens (strings) and token_ids (integers),
keeps track of token frequencies, and can yield samples from the
unigram distribution.
'''

import os
from token_map import TokenMap, SILENT, WARN, ERROR, UNK
from counter_sampler import CounterSampler

class UnigramDictionary(object):
    '''
    Bundles together a TokenMap and CounterSampler.  Provides a method for
    pruning the vocabluary while keeping the TokenMap and CounterSampler
    in sync with one another.
    '''


    def __init__(self, on_unk=WARN, token_map=None, counter_sampler=None):
        '''
        Create a new UnigramDictionary.  Typical usage provides no
        arguments, but a token_map and counter_sampler can be provided
        to build a UnigramDictionary that comprises them.
        '''
        self.on_unk = on_unk
        self.token_map = token_map
        if token_map is None:
            self.token_map = TokenMap(on_unk=on_unk)

        self.counter_sampler = counter_sampler
        if counter_sampler is None:
            self.counter_sampler = CounterSampler()


    def __contains__(self, token):
        return token in self.token_map.map


    def sort(self):
        unk_count = self.counter_sampler.counts[0]

        # Get the counts and tokens (skipping the first UNK entry)
        # They are parallel arrays (ith count corresponds to ith token)
        counts = self.counter_sampler.counts[1:]
        tokens = self.token_map.tokens[1:]

        # Zip them together and sort by counts
        token_counts = zip(counts, tokens)
        token_counts.sort(reverse=True)

        # Separate them again
        new_counts = [unk_count]
        new_tokens = ['UNK']
        for count, token in token_counts:
            new_counts.append(count)
            new_tokens.append(token)

        # Rebuild the token_map and counter_sampler on the sorted arrays
        self.token_map = TokenMap(on_unk=self.on_unk, tokens=new_tokens)
        self.counter_sampler = CounterSampler(counts=new_counts)


    def remove(self, token):
        idx = self.get_id(token)
        self.token_map.remove(token)
        self.counter_sampler.remove(idx)


    def compact(self):
        self.token_map.compact()
        self.counter_sampler.compact()


    def prune(self, min_frequency=5):
        '''
        Remove all tokens that have been observed fewer than min_frequency
        times.  Counts for tokens that are removed are attributed to UNK.
        '''
        counts = []
        tokens = []
        discarded = set()
        for idx, token in enumerate(self.token_map.tokens):

            # Copy over tokens that have at least min_frequency
            # observations. Also copy over UNK no matter what it's
            # frequency.
            if (
                self.counter_sampler.get_frequency(idx) >= min_frequency
                or idx == 0
            ):
                tokens.append(token)
                counts.append(self.get_frequency(idx))

            # Skip tokens that have too little frequency.  Attribute their
            # observations to UNK
            else:
                counts[UNK] += self.get_frequency(idx)
                discarded.add(token)

        # Create a new TokenMap and CounterFrequency based on the
        # filtered tokens and their counts
        self.token_map = TokenMap(on_unk=self.on_unk, tokens=tokens)
        self.counter_sampler = CounterSampler(counts=counts)

        return discarded


    def add(self, token):
        '''
        Add a new token.  If this "token type" (which means this specific
        spelling of a word) has not been seen before, add it to the
        mapping.  Also increment the count for that token type.  Return
        its ID under the token mapping.
        '''

        # Get or create an id for this token
        token_id = self.token_map.add(token)

        # Increment the frequency count
        self.counter_sampler.add(token_id)

        return token_id


    def add_count(self, token, count):
        '''
        Add `count` to the counts for `token`, making a new entry if 
        necessary.
        '''
        # Get or create an id for this token
        token_id = self.token_map.add(token)
        # Increment the frequency count
        self.counter_sampler.add_count(token_id, count)


    def get_vocab_size(self):
        '''
        Return the number of unique tokens in the token_map.
        '''
        return len(self.token_map)


    def get_num_tokens(self):
        '''
        Return the total number of (non-distinct) tokens observed.
        '''
        return len(self.counter_sampler)


    def __len__(self):
        '''
        Same as get_vocab_size().
        Return the number of unique tokens in the token_map.
        '''
        return len(self.token_map)


    def update(self, token_iterable):
        '''
        Like `add`, but accepts an iterable of tokens, incrementing the
        count for each of them.
        '''
        return [self.add(token) for token in token_iterable]


    def add_dictionary(self, other):
        '''
        Adds counts from another UnigramDictionary, `other`, to `self`'s
        counts, i.e. adding in place.
        '''
        self.update_counts(other.get_frequency_list())


    def update_counts(self, token_counts_iterable):
        '''
        Like `add_count` but accepts an iterable of (token,count) pairs,
        and increments the count for each token by the count given.
        Expected usage is to have a dictionary with tokens as keys
        and counts as values, and pass in your_dict.iteritems().
        '''
        return [
            self.add_count(token, count) 
            for token, count in token_counts_iterable
        ]


    def get_id(self, token):
        '''
        Get the id (int) for the corresponding token (string).
        '''
        # Delegate to the underlying token_map.
        return self.token_map.get_id(token)


    def get_ids(self, token_iterable):
        '''
        Get the ids (list of ints) for the corresponding tokens (strings)
        issued by token_iterable.
        '''
        # Delegate to the underlying token map.
        return self.token_map.get_ids(token_iterable)


    def get_token(self, idx):
        '''
        Return token (string) for the corresponding id (int)
        '''
        # Delegate to the underlying token map
        return self.token_map.get_token(idx)


    def get_tokens(self, idx_iterable):
        '''
        Return the tokens (list of strings) for the corresponding ids
        (ints) issued by idx_iterable.
        '''
        # Delegate to the underlying token map.
        return self.token_map.get_tokens(idx_iterable)


    def save(self, savedir):
        '''
        Save the UnigramDictionary to the directory specified.  This saves
        the underlying TokenMap and CounterSampler in the directory
        given (savedir), using the default filenames "token-map.gz" and
        "counter-sampler.gz".
        '''

        # If the directory provided is a file, raise an error
        if os.path.exists(savedir):
            if os.path.isfile(savedir):
                raise IOError(
                    'Directory specified for saving UnigramDictionary is a '
                    'file.'
                )

        # If the directory provided doesn't exist, make it (this will not
        # make parent directories though).
        else:
            os.mkdir(savedir)


        # Save the TokenMap and CounterSampler by delegating to their
        # save functions.
        self.token_map.save(os.path.join(savedir, 'token-map.gz'))
        self.counter_sampler.save(os.path.join(
            savedir, 'counter-sampler.gz'
        ))


    def load(self, loaddir):
        '''
        Load a UnigramDictionary from the specified directory, by
        loading the TokenMap and CounterSampler stored there.  This assumes
        the filenames are 'token-map.gz' and 'counter-sampler.gz'.
        '''
        # Load the TokenMap by delegation to its load function
        self.token_map = TokenMap()
        self.token_map.load(os.path.join(loaddir, 'token-map.gz'))

        # Load the CounterSampler by delegation to its load function
        self.counter_sampler = CounterSampler()
        self.counter_sampler.load(
            os.path.join(loaddir, 'counter-sampler.gz'))


    def get_token_list(self):
        '''
        Gets an iterable of tokens currently in the dictionary.  Omits
        The 'UNK' token.
        '''
        return (
            token for token in self.token_map.tokens if token is not 'UNK'
        )


    def get_frequency_list(self):
        '''
        Gets an iterable of (token, count) tuples.
        '''

        # Handle the case where there are no counts at all yet
        if len(self.counter_sampler.counts) == 0:
            return []

        # Otherwise get the counts normally
        return (
            (token, self.get_frequency(self.get_id(token)))
            for token in self.token_map.tokens
        )


    def sample(self, shape=None):
        '''
        Draw a sample according to the counter_sampler probability
        '''
        # Delegate to the underlying CounterSampler
        return self.counter_sampler.sample(shape)


    def get_probability(self, token_id):
        '''
        Return the probability associated to token_id.
        '''
        # Delegate to the underlying CounterSampler
        return self.counter_sampler.get_probability(token_id)


    def get_token_frequency(self, token):
        '''
        Return the frequency (count) associated to the token
        '''
        token_id = self.get_id(token)
        # If the token is unknown, return 0
        if token_id == UNK:
            return 0
        return self.get_frequency(token_id)


    def get_frequency(self, token_id):
        '''
        Return the frequency associated to token_id.
        '''
        # Delegate to the underlying CounterSampler
        return self.counter_sampler.get_frequency(token_id)




from xml.dom import minidom

# This provisional dom is used as an element factory
DOM = minidom.Document()

def element(tag_name, attributes={}):
    elm = DOM.createElement(tag_name)
    bind_attributes(elm, attributes)
    return elm 

def bind_attributes(element, attributes):
    for attribute in attributes:
        element.setAttribute(attribute, attributes[attribute])
    return element

def div(attributes={}):
    return element('div', attributes)

def span(attributes={}):
    return element('span', attributes)

def text(text_content):
    return DOM.createTextNode(str(text_content))

def table(attributes={}):
    return element('table', attributes)

def tr(attributes={}):
    return element('tr', attributes)

def td(attributes={}):
    return element('td', attributes)


def build_table(fields):
    table_elm = table({'class': 'performance'})
    first_row = True
    for row in fields:
        if first_row:
            tr_elm = table_elm.appendChild(tr({'class': 'first-row'}))
            first_row = False
        else:
            tr_elm = table_elm.appendChild(tr())
                
        first_cell = True
        for cell in row:
            if first_cell:
                td_elm = tr_elm.appendChild(td({'class': 'first-cell'}))
                first_cell = False
            else:
                td_elm = tr_elm.appendChild(td())
            td_elm.appendChild(text(cell))

    return table_elm


class Styler(dict):

    def as_element(self):
        style_element = DOM.createElement('style')
        style_element.appendChild(text('\n' + self.serialize()))
        return style_element

    def serialize(self):
        string = ''
        for rule in self:

            # Open the style rule
            string += '%s {\n' % rule

            # Write each directive
            for directive in self[rule]:
                string += '\t%s: %s;\n' % (directive, self[rule][directive])

            # Close the style rule
            string += '}\n'

        return string



import traceback
import sys


def trace():
	'''
		When called in an except block, returns the stack trace that
		would have been emmitted due to the exception if it were uncaught.
	'''
	return traceback.format_exc(sys.exc_info())


def get_trace():
	'''
		Similar to get_trace, but returns an array of strings, each one
		matching a stack frame.  See trace().
	'''
	# get the stack trace.  
	trc = traceback.format_stack()
	trc =  trc[1:-1]
	trc.append(traceback.format_exc())

	return trc




#!/usr/bin/env python

import shutil
import multiprocessing
from multiprocessing.managers import BaseManager
import time
import random
import t4k

def run_test1():
	tracker = t4k.SharedProgressTracker('test_dict')

	p1 = multiprocessing.Process(
			target=test_concurrent_write, args=(tracker,'A'))
	p2 = multiprocessing.Process(
			target=test_concurrent_write, args=(tracker,'B'))

	p1.start()
	p2.start()

	p1.join()
	p2.join()

	print tracker['yo'], '(should be 20000.)'
	tracker.close()


def run_test2():
	tracker = t4k.SharedProgressTracker('test_dict')

	p1 = multiprocessing.Process(target=test_hold, args=(tracker,'A'))
	p2 = multiprocessing.Process(target=test_hold, args=(tracker,'B'))

	start_time = time.time()

	p1.start()
	p2.start()

	p1.join()
	p2.join()

	print [int(tracker['A:'+str(i)] - start_time) for i in range(10)]
	print [int(tracker['B:'+str(i)] - start_time) for i in range(10)]
	print (
		'(numbers in second list should all be greater than or equal'
		' to those in first list.)'
	)

	tracker.close()


def test_hold(tracker, name):
	tracker.hold()
	for i in range(10):
		print '+'
		time.sleep(random.random())
		tracker['%s:%d' % (name, i)] = time.time()
	tracker.unhold()


def test_concurrent_write(tracker, name):
	for i in range(10000):
		if i % 1000 == 0:
			print '.'
		tracker.hold()
		if 'yo' in tracker:
			tracker['yo'] += 1
		else:
			tracker['yo'] = 1
		tracker.unhold()





# Tried doing this using a manager

POD_PUBLIC_METHODS = [
	'hold',
	'unhold',
	'keys',
	'values',
	'read',
	'mark_dirty',
	'sync',
	'escape_key',
	'__iter__',
	'next',
	'__contains__',
	'__len__',
	'__getitem__',
	'ensure_unicode',
	'update',
	'__setitem__',
	'check_or_add',
	'set_item',
	'set',
	'check',
	'add',
	'increment_tries',
	'reset_tries',
	'mark_done',
	'mark_not_done',
]


def test_manager():

	class MyManager(BaseManager):
		pass

	MyManager.register(
		'ProgressTracker', t4k.ProgressTracker, exposed=POD_PUBLIC_METHODS
	)
	manager = MyManager()
	manager.start()

	tracker = manager.ProgressTracker('test_dict')

	p1 = multiprocessing.Process(
		target=test_concurrent_write, args=(tracker,'A'))
	p2 = multiprocessing.Process(
		target=test_concurrent_write, args=(tracker,'B'))

	p1.start()
	p2.start()

	p1.join()
	p2.join()

	print tracker['yo']


if __name__=='__main__':
	# Remove pre-existing left-over test files from previous test
	t4k.ensure_removed('test_dict')

	# Run tests
	#test_manager()
	#run_test2()
	run_test1()



'''
Calculates the maximum likelihood probability of single words, as equal
to their frequency within a corpus, and enables efficiently sampling
from the distribution.

Assumes that the words have been converted into ids that were assigned
by auto-incrementing from 0 (The consequence if that is not the case isn't dire,
it just means memory would be wasted)
'''

import numpy as np
import gzip


class CounterSamplerException(Exception):
	pass


class CounterSampler(object):

	def __init__(self, counts=[]):
		'''
		Create a CounterSampler.  Most common usage is to provide no
		arguments, creating an empty CounterSampler, because CounterSampler
		provides functions to accumulate counts from observations.

		counts:	list of counts.  The index of the count in the counts list
			serves as the id for the outcome having that many counts.
		'''

		# Validation
		if not all([isinstance(c, int) for c in counts]):
			raise ValueError('Counts must be an iterable of integers')
		self.counts = list(counts)
		self.sampler_ready = False


	def remove(self, idx):
		displaced_count = self.counts[idx]
		self.counts[idx] = None
		self.sampler_ready = False


	def compact(self):
		'''
		Remake the conts list, eliminating `None`s which are holes
		left by calls to `remvoe()`.
		'''
		self.counts = [c for c in self.counts if c is not None]


	def pad_if_necessary(self, idx):
		if len(self.counts) <= idx:
			add_zeros = [0] * (idx +1 - len(self.counts))
			self.counts.extend(add_zeros)



	def add(self, idx):
		'''
		Increment the count for `idx` by 1.
		'''
		self.sampler_ready = False
		self.pad_if_necessary(idx)
		self.counts[idx] += 1


	def add_count(self, idx, count):
		'''
		Like `self.add()`, but increment the count for `idx` by `count`
		rather than just 1
		'''
		self.sampler_ready = False
		self.pad_if_necessary(idx)
		self.counts[idx] += count

	def update(self, idxs):
		for idx in idxs:
			self.add(idx)


	def ensure_prepared(self):

		if self.sampler_ready:
			return

		if len(self.counts) < 1:
			raise CounterSamplerException(
				'Cannot sample if no counts have been made'
			)

		if not self.sampler_ready:
			self.total = np.sum(self.counts)
			self.probabilities = np.array(self.counts) / float(self.total)
			self.sampler_ready = True


	def get_total_counts(self):
		self.ensure_prepared()
		return self.total


	def __len__(self):
		self.ensure_prepared()
		return self.total


	num_to_load = 10**5
	def sample(self, shape=None):

		if shape is not None:
			size = np.prod(shape)
		else:
			size = 1

		if not hasattr(self, '_probs'):
			self._probs = np.array(self.counts, dtype='float64')
			self._probs = self._probs / np.sum(self._probs)

		if not hasattr(self, '_np_sample'):
			num_to_load = max(self.num_to_load, 2*size)
			self._np_sample = np.random.choice(
				a=len(self._probs), size=num_to_load, p=self._probs
			)

		if not hasattr(self, '_ptr'):
			self._ptr = 0

		self._ptr += size

		if self._ptr >= len(self._np_sample):
			num_to_load = max(self.num_to_load, 2*size)
			self._np_sample = np.random.choice(
				a=len(self._probs), size=num_to_load, p=self._probs
			)
			self._ptr = 0
			return self.sample(shape)

		else:
			if shape is not None:
				return self._np_sample[self._ptr - size: self._ptr].reshape(shape)
			else:
				return self._np_sample[self._ptr - size: self._ptr][0]


	def save(self, filename):
		if filename.endswith('.gz'):
			f = gzip.open(filename, 'w')
		else:
			f = open(filename, 'w')

		for c in self.counts:
			f.write('%d\n' % c)


	def load(self, filename):
		if filename.endswith('.gz'):
			f = gzip.open(filename)
		else:
			f = open(filename)

		self.counts = [int(c) for c in f.readlines()]


	def get_frequency(self, idx):
		'''
		Return the number of observations for outcome idx.
		'''
		return self.counts[idx]


	def get_probability(self, token_id):
		'''
		Return the probability associated to token_id.
		'''
		self.ensure_prepared()
		return self.probabilities[token_id]



from natsort import humansorted
import shutil
import re
import subprocess
import os


def file_empty(file_path):
    return os.stat(file_path).st_size == 0

def ensure_removed(path):
    """
    Ensure that the given path does not exist, by removing it if necessary.
    """
    if os.path.exists(path):
        shutil.rmtree(path)    

def ensure_exists(path):
    """
    Ensure that the given path exists, by creating it if it doesn't.  If the
    path exists but is a file, an `OSError` would be raised.
    """
    if not os.path.exists(path):
        os.makedirs(path)    


def ls(
    path,
    files=True,
    dirs=True,
    match=None,
    exclude=None,
    absolute=False,
    relative='.',
    basename=False,
    recurse=False,
    list_all=False,
    natural_sort=True,
    iteritems=False
):

    lister = PathLister(
        path=path,
        files=files,
        dirs=dirs,
        match=match,
        exclude=exclude,
        absolute=absolute,
        relative=relative,
        basename=basename,
        recurse=recurse,
        list_all=list_all,
        natural_sort=natural_sort,
        iteritems=iteritems
    )
    return lister.generate()


ABSOLUTE = 0
RELATIVE = 1
BASENAME = 2
class PathLister(object):
    '''
        Iterator for listing all files in a directory.
        - Lists files found under path, unless <files> is False
        - Can also list directories, if <dirs> is True
        - Only files (and directories) matching <match> 
            will be returned, unless they also match <exclude>.
        - By default, doesn't descend into subdirectories, unless
            <recurse> is True, lists files in subdirs too.
        - If <recurse> is True, only directories matching <match>
            and not matching <exclude> will be followed
        - Paths to files are returned relative to the current working 
            directory, unless <absolute> is true, then absolute paths are 
            returned, unless basename is true, then only the name of the 
            file or directory is returned.
        - Files are listed in natural sort ordering (i.e. respecting
            numerical order), but if <natural_sort> is False alphabetical
            order is used instead.
        - Hidden files (starting with '.') are ommitted, but if <list_all>
            is true, then they will be included.  In recursive mode, this
            is necessary in order to follow hidden folders.
    '''

    def __init__(
        self,
        path,
        files=True,
        dirs=True,
        match=None,
        exclude=None,
        absolute=False,
        relative='.',
        basename=False,
        recurse=False,
        list_all=True,
        natural_sort=True,
        iteritems=False
    ):
        self.path = path
        self.files = files
        self.dirs = dirs
        self.match =  match and re.compile(match)
        self.exclude = exclude and re.compile(exclude)
        self.absolute = absolute
        self.relative = relative
        self.basename = basename
        self.recurse = recurse
        self.list_all = list_all
        self.natural_sort = natural_sort
        self.iteritems=iteritems

        # Validation -- make sure self.path exists
        if not os.path.exists(self.path):
            raise OSError('no such file or directory: %s' % self.path)


    def filter_back_pointers(self, dirs):
        '''
            Exclude '.' and '..' from <dirs>.
        '''
        return filter(
            lambda i: not (i.split('/')[-1]=='.' or i.split('/')[-1]=='..'),
            dirs
        )


    def generate(self):
        '''
        Either return the path generator, or a list of paths, depending on 
        the constructor.
        '''
        self.walker = os.walk(self.path)
        self.next_dir()
        if self.iteritems:
            return self._generate()
        else:
            return [item for item in self._generate()]


    def next_dir(self):
        cur_path, cur_dirs, cur_files = self.walker.next()

        # First reletavize cur_path to current working directory (by default
        # it's reletavized to the path given in the constructor)
        cur_path = os.path.relpath(cur_path, '.')

        # Do we want the files, the directories, both?
        if self.files:
            self.items = cur_files
        else:
            self.items = []
        if self.dirs:
            self.items.extend(cur_dirs)


        # The walker has yielded basenames for the files and dirs. 
        # If that's what we want, then don't modify it
        if self.basename: 
            pass

        # But if we want absolute paths, the absolutize them now.
        elif self.absolute:
            self.items = [
                os.path.abspath(os.path.join(cur_path, d)) 
                for d in self.items
            ]

        # Otherwise reletavize to the locatin of self.relative, which is by 
        # default the current working directory.
        elif cur_path != '.':
            self.items = [
                os.path.relpath(os.path.join(cur_path, f), self.relative)
                for f in self.items
            ]

        # Sort the items either "naturally" or alphabetically
        if self.natural_sort:
            self.items = humansorted(self.items, reverse=True)
        else:
            self.items.sort(reverse=True)

        # Filter items against the exclude and match if any
        if self.match is not None:
            self.items = [i for i in self.items if self.match.search(i)]
        if self.exclude is not None:
            self.items = [
                i for i in self.items if not self.exclude.search(i)]


    def _generate(self):
        '''
            yield the next item, and absolutize the path if necessary.
        '''
        # We'll break out of this loop when stopIteration is raised by 
        # the call to self.get_next()
        while True:
            yield self.get_next()


    def get_next(self):

        while True:

            # Try popping off the next item
            try:
                return self.items.pop()
            except IndexError:
                pass

            # If there's no more items, and if we're in recursive mode, 
            # descend into the next directory
            if self.recurse:
                self.next_dir()
                return self.get_next()

            # And if we're not in recursive mode, then we're done
            else:
                raise StopIteration


    def _ls(self, path, filter_back_pointers=True):

        # Ask OS to list the files and directories in path

        #if self.list_all:
        #    items = check_output(['ls', '-a', path]).split()
        #else:
        #    items = check_output(['ls', path]).split()

        ls = subprocess.Popen(list_command, stdout=subprocess.PIPE)


        # Sort the files and directories
        if self.natural_sort:
            sort_command = ['sort', '-n', '-r']
        else:
            sort_command = ['sort', '-r']
        items = subprocess.check_output(sort_command, stdin=ls.stdout)
        ls.wait()
        items = [os.path.join(path, i) for i in items.split()]

        # Get rid of '.' and '..', if necessary
        if self.list_all and filter_back_pointers:
            items = self.filter_back_pointers(items)

        # Go through the list separate files from directories
        files = []
        dirs = []
        for item in items:

            # Skip if it doesn't match match, or does match exclude
            whitelist_match = (
                True if self.match is None 
                else self.match.search(item) 
            )
            blacklist_match = (
                false if self.exclude is None
                else self.exclude.search(item)
            )
            if not whitelist_match or blacklist_match:
                continue

            # Append files to the list of files
            if os.path.isfile(item):
                files.append(item)

            # Append directories to the list of dirs
            else:
                dirs.append(item)

        return files, dirs




# -*- coding: utf-8 -*-
import re
import nltk


class Vectorizer(object):

	SPLITTING_PATTERN = re.compile(
		#r'(?:\s|!|,|—|\(|\)|\[|\]|<|>|\'|"|`|~|/|\\|\?|\||:|;|{|})+'
		r'(?:\s|!|,|—|\(|\)|\[|\]|<|>|"|`|~|\\|\?|\||;|{|})+'
	)
	TERMINAL_PUNCTUATION = r'\.|-|\''
	LEADING_PUNCTUATION = re.compile('^(' + TERMINAL_PUNCTUATION + ')*')
	TRAILING_PUNCTUATION = re.compile( '(' + TERMINAL_PUNCTUATION + ')*$')

	def __init__(self, do_lemmatize=True):
		self.vocab = {}
		self.lookup = []
		self.do_lemmatize = do_lemmatize
		self.lemmatizer = nltk.stem.wordnet.WordNetLemmatizer()


	def get_size(self):
		return len(self.lookup)


	def add(self, token):
		token = self.clean(token)
		if token in self.vocab:
			raise ValueError(
				'The token "%s" is already in the vocabulary.' % token
			)

		self.vocab[token] = len(self.lookup)
		self.lookup.append(token)


	def add_doc(self, doc):
		tokens = self.SPLITTING_PATTERN.split(doc)
		for t in tokens:
			self.add(t)


	def clean(self, token):
		token = token.strip()

		if self.looks_like_url(token):
			return token

		token = self.LEADING_PUNCTUATION.sub('', token)
		token = self.TRAILING_PUNCTUATION.sub('', token)
		if self.do_lemmatize:
			token = self.lemmatizer.lemmatize(token)

		token = token.strip()

		return token


	def looks_like_url(self, token):
		return token.startswith('http')


	def code_to_token(self, code):
		try:
			return self.lookup[code]
		except IndexError:
			raise ValueError(
				'There is no entry in the vocabulary at code %d. '
				'The vocabulary size is only %d.' % (code, len(self.lookup))
			)


	def token_to_code(self, token):
		cleaned_token = self.clean(token)
		try:
			return self.vocab[cleaned_token]
		except KeyError:
			self.add(token)
			return self.vocab[cleaned_token]


	def doc_to_vec(self, doc):
		vec = {}
		tokens = self.SPLITTING_PATTERN.split(doc)
		for token in tokens:
			code = self.token_to_code(token)
			try:
				vec[code] += 1
			except KeyError:
				vec[code] = 1

		return vec








import random

class ReservoirSampler(object):

    def __init__(self, k):
        self.sample = []
        self.k = k
        self.i = 0


    def add(self, item):
        """
        Possibly add an item to the sample.  The probability that an item will
        be kept is continually decreased as more items are considered such, at
        any given time, all items that have been considered so far have an
        equal chance of being part of the set.
        """

        # Increment the counter for the number of items considered
        i = self.i
        self.i += 1

        # If we don't yet have k samples, then we definitely include the item
        if i < self.k:
            self.sample.append(item)
            return True

        # Otherwise, we *might* keep the item (replacing a previously kept
        # item).  The probability adjusts as we go along so that all items are
        # given uniform chance of being in the final set if retained items
        else:
            pick = random.randint(0, i)
            if pick < self.k:
                self.sample[pick] = item
                return True

        return False


    def add_many(self, iterable, allow_small_sample=False):
        """
        Consider each element yielded by the iterable for inclusion in the
        sample.  If allow_small_sample is False, then, if fewer than self.k
        items are in the sample after the iterable is exhausted, an exception
        will be raised.
        """
        for item in iterable:
            self.add(item)

        if len(self.sample) < self.k and not allow_small_sample:
            raise ValueError('Population smaller than sample size')


def reservoir_sample(iterable, k, allow_small_sample=False):
    """
    Convenience function to retrieve a reservior sample from an iterable
    """
    sampler = ReservoirSampler(k)
    sampler.add_many(iterable, allow_small_sample)
    return sampler.sample





#!/usr/bin/env python
from setuptools import setup

setup(
	name='t4k',
	url='http://drgitlab.cs.mcgill.ca/enewe101/trix_for_kids',
	version='0.6.2',
	description='my personal swiss army knife',
	author='edward newell',
	author_email='edward.newell@gmail.com',
	packages=['t4k'],
	license='MIT',
	classifiers=[
		'Development Status :: 2 - Pre-Alpha',
		'Intended Audience :: Science/Research',
		'Intended Audience :: Information Technology',
		'Programming Language :: Python :: 2.7',
		'Natural Language :: English',
		'Topic :: Utilities',
	],
	keywords='trix are for kids',
	install_requires=[
        'numpy', 'psutil', 'natsort', 'Pyro4', 'setproctitle', 'tastypy'
    ],
    scripts=['bin/mem-server']
)




import numpy as np
import random
import math
import hashlib


def skipfirst(iterable):
    iterator = iter(iterable)
    iterator.next()
    return iterator


def get_fold(data, num_folds, this_fold):
    """
    Randomly, but reproducibly separate out 1/num_folds of ``data`` as a
    validation set.

    By varying ``this_fold`` from 0 to ``num_folds - 1``, obtain different
    splits of the data such that the validations sets do not overlap and all
    data appear in some validation set.
    """
    if this_fold < 0 or this_fold >= num_folds:
        raise ValueError('``this_fold`` must be in [0,num_folds-1]')

    # Reproducibly create a random permutation
    random.seed(0)
    permutation = range(len(data))
    random.shuffle(permutation)

    # Get the slice of the random permutation that will correspond to the test
    # part of the fold
    start = int(this_fold * float(len(data)) / num_folds)
    end = int((this_fold+1) * float(len(data)) / num_folds)
    fold_test_indices = permutation[start:end]

    # Everything else is the training part of the fold
    fold_train_indices = permutation[:start] + permutation[end:]

    fold_test = np.array([data[i] for i in fold_test_indices])
    fold_train = np.array([data[i] for i in fold_train_indices])
    return fold_train, fold_test




class IncrementingMap(dict):

    def add(self, key):
        if key not in self:
            self[key] = self.get_incrementing_id()
            self._get_keys().append(key)


    def add_many(self, keys_iterator):
        for key in keys_iterator:
            self.add(key)


    def get_incrementing_id(self):
        try:
            self._current_id += 1
        except AttributeError:
            self._current_id = 0
        return self._current_id


    def _get_keys(self):
        '''
        This getter is private, and covers the case where self._keys is not 
        yet defined.  The first time it is called, self._keys is initialized
        to be an empty list.
        '''
        try:
            return self._keys
        except AttributeError:
            self._keys = []
            return self._keys


    def key(self, idx):
        return self._get_keys()[idx]


    def keys(self):
        return [k for k in self._get_keys()]



def _validate_normalize_slice_indices(list_obj, start, stop):

    # Fail if list_item doesn't support indexing
    if not hasattr(list_obj, '__getitem__'):
        raise ValueError('Cannot index into list_obj.')

    # Fail if list_item doesn't support indexing
    if not hasattr(list_obj, '__getitem__'):
        raise ValueError('Cannot index into list_obj.')

    # If stop is not defined, set it equal to len(list_obj)
    the_len = len(list_obj)
    if stop is None:
        stop = the_len

    # Convert negative indices into equivalent positive form
    if start < 0:
        start = start + the_len
    if stop < 0:
        stop = stop + the_len

    # trim start and stop to within the list's actual length
    start = max(0, start)
    stop = min(the_len, stop)

    ## Fail if start is greater than stop
    #if start > stop:
    #       raise ValueError('start cannot be greater than stop.')

    return start, stop


def ltrim(list_obj, item=0):
    match = lambda x: x!=item
    start = lindex(list_obj,match_func=match) 
    return list_obj[start:]

def rtrim(list_obj, item=0):
    match = lambda x: x!=item
    stop = rindex(list_obj,match_func=match)+1
    return list_obj[:stop]

def trim(list_obj, item=0):
    match = lambda x: x!=item
    start = lindex(list_obj,match_func=match) 
    stop = rindex(list_obj,match_func=match)+1
    return list_obj[start:stop]


def lindex(list_obj, item=None, match_func=None, start=0, stop=None):

    ''' 
    find the index of first occurence of <item> in <list_obj>, within 
    the slice defined by <start> and <stop>.  Start and stop work
    the way normal list slicing does.  By default, start is 0 and stop
    is the length of the list.  If stop is not given, <list_obj> must
    define __len__().
    '''

    start, stop = _validate_normalize_slice_indices(list_obj, start, stop)

    if match_func == None:
        if item is None:
            raise ValueError(
                'Either `item` or `match_func` must be specified')
        match_func = lambda x: x==item

    # find the index of item
    cur_idx = start
    while cur_idx < stop:
        try:
            if match_func(list_obj[cur_idx]):
                return cur_idx
        except IndexError:
            pass
        cur_idx += 1

    # Item is not in the list
    raise ValueError('No match in list.')



def rindex(list_obj, item=None, match_func=None, start=0, stop=None):

    ''' 
    find the index of first occurence of <item> in <list_obj>, within 
    the slice defined by <start> and <stop>.  Start and stop work
    the way normal list slicing does.  By default, start is 0 and stop
    is the length of the list.  If stop is not given, <list_obj> must
    define __len__().
    '''

    start, stop = _validate_normalize_slice_indices(list_obj, start, stop)

    if match_func == None:
        if item is None:
            raise ValueError(
                'Either `item` or `match_func` must be specified')
        match_func = lambda x: x==item

    # find the index of item
    cur_idx = stop - 1
    while cur_idx >= start:
        try:
            if match_func(list_obj[cur_idx]):
                return cur_idx
        except IndexError:
            pass
        cur_idx -= 1

    # Item is not in the list
    raise ValueError('No match in list')


def indices(list_obj, item, start=0, stop=None):
    idxs = []

    try:
        while True:
            idx = lindex(list_obj, item, start, stop)
            idxs.append(idx)
            start = idx + 1
    except ValueError:
        pass

    return idxs
        

def skip(list_obj, skip_slices):
    '''
    Returns a copy of list obj with elements ommited.  The ommited elements
    are specified by skip_slices, a list of tuples specifying spans of
    elements to skip using slice notation.

    -- Input --
    skip_slices: a list of 2-tuples, each consisting of start and stop
        indices specifying spans of elements to be ommitted from list_obj
        
    list_obj: a list.
    
    -- Returns --
    censored_list: a list that doesn't contain the elements that
        would be selected by treating the tuples in skip_slices as
        slice indices.
    '''

    # We begin by normalizing the slice indices.  First, convert all
    # occurrences of None into integers.  
    skip_slices_no_none = []
    for start, stop in skip_slices:

        # When None is used as a start index, it is equivalent to 0
        if start is None:
            start = 0

        # When None is used as a stop index, it is equivalent to len(list)
        if stop is None:
            stop = len(list_obj)

        # Append the equivalent slice defined using only integers
        skip_slices_no_none.append((start,stop))
    
    # To assemble the 
    # censored list, we need the slice indices sorted by start-index.
    # We also need to handle cases where the spans defined by slice
    # indices overlap.  When slice indices overlap, there are two cases.
    # In the simple case, the spans overlap a little, but in the second
    # case, the first span subsumes the second.  We need to eliminate
    # occurrences of the second kind.  We can simply drop the subsumed
    # span since it is redundant anyway.
    normalized_skip_slices = []
    for i, (start, stop) in enumerate(sorted(skip_slices_no_none)):

        # On the first iteration, just keep the first slice and move on
        if i == 0:
            normalized_skip_slices.append((start,stop))
            prev_start, prev_stop = start, stop

        # This means that the current slice is subsumed by the previous.
        # We simply don't include it
        elif prev_stop > stop:
            pass

        # The normal case: prev span doesn't subsume current span, so it
        # is safe to add it.
        else:
            normalized_skip_slices.append((start,stop))
            prev_start, prev_stop = start, stop

    # Now that the skip slices are sorted and there are no subsumed
    # spans, we proceed as follows.  Start with an empty list, and 
    # for each span, add the part of the list which comes before the
    # current span's start but after the previous span's stop
    censored_list = []
    prev_stop = None
    for start, stop in normalized_skip_slices:
        censored_list.extend(list_obj[prev_stop:start])
        prev_stop = stop

    censored_list.extend(list_obj[prev_stop:None])

    return censored_list







def flatten(iterable, recurse=False, depth=0):
    '''
    flattens an iterable of iterables into a simple list.  E.g.
    turns a list of list of elements into a simple list of elements.

    <recurse> will attempt to flatten the elements within the inner 
    iterable too, and will continue so long as the elements found are 
    iterable.  However, recurse will treat strings as atomic, even though
    they are iterable.
    '''

    flat_list = []
    for element in iterable:

        # If the element is not iterable, just add it to the flat list
        try:
            iter(element)
        except:
            flat_list.append(element)
            continue
        else:

            # If the element is a string, treat it as not iterable
            if isinstance(element, basestring):
                flat_list.append(element)

            # If the element is iterable and <recurse> is True, recurse
            elif recurse:
                flat_list.extend(flatten(element, recurse))

            # If the element is iterable but <recurse> is False, extend
            else:
                flat_list.extend(element)

    return flat_list



def group(iterable, num_chunks):
    '''
    Breaks <iterable> into <num_chunks> chunks (lists) of approximately 
    equal size (chunks will differ by at most one item when <num_chunks> 
    doesn't evenly divide len(<iterable>)).
    '''

    # Coerce the sequence into a list
    iterable = list(iterable)

    # Figure out the chunksize
    num_items = len(iterable)
    chunk_size = num_items / float(num_chunks)

    # Allocate approximately chunksize elements to each chunk, rounded off.
    chunks = []
    for i in range(num_chunks):
        start_index = int(math.ceil(i * chunk_size))
        end_index = int(math.ceil((i+1) * chunk_size))
        chunks.append(iterable[start_index:end_index])

    return chunks


def chunk(iterable, chunk_size):
    '''
    Returns a list of lists of items from iterable, where the internal 
    lists are each approximately of size chunk_size (except the last one,
    which is smaller if chunk_size doesn't evenly divide len(iterable)
    '''

    iterator = iter(iterable)
    chunks = []
    this_chunk = []
    still_has_items = True
    while still_has_items:
        try:
            this_chunk.append(iterator.next())
        except StopIteration:
            still_has_items = False

        if len(this_chunk) == chunk_size:
            chunks.append(this_chunk)
            this_chunk = []

        elif not still_has_items and len(this_chunk) > 0:
            chunks.append(this_chunk)

    return chunks


def rangify(iterable):
    '''
    Converts a list of indices into a list of ranges (compatible with the
    range function).  E.g. [1,2,3,7,8,9] becomes [(1,4),(7,10)].
    The iterable provided must be sorted. Duplicate indices are ignored.  
    '''
    ranges = []
    last_idx = None
    for idx in iterable:

        if last_idx is None:
            start = idx

        elif idx - last_idx > 1:
            ranges.append((start, last_idx + 1))
            start = idx

        last_idx = idx

    if last_idx is not None:
        type(ranges)
        ranges.append((start, last_idx + 1))

    return ranges




def binify(string_id, num_bins):
    ''' 
    Uniformly assign objects to one of `num_bins` bins based on the
    hash of their unique id string.
    '''
    hexdigest = hashlib.sha1(string_id).hexdigest()
    return int(hexdigest,16) % num_bins


def inbin(string_id, num_bins, this_bin):
    if this_bin >= num_bins:
        raise ValueError(
            '`this_bin` must be an integer from 0 to `num_bins`-1.')
    return binify(string_id, num_bins) == this_bin




'''
The persistent ordered dict is a data struture that can be interacted with
like a python dict, and which is easy to keep synced with an on-disk copy

This allows you to easily persist data between non-concurrent runs of a
program.  It's useful for keeping track of progress in long jobs.
'''

import math
import atexit
import signal
import time
import sys
import multiprocessing
import subprocess
import json
import os
import re
import copy
from file_utils import ls
import gzip
import tastypy

LINES_PER_FILE = 1000


# Define Exceptions for the PersistentOrderedDict class and its subclasses
class PersistentOrderedDictException(Exception):
    pass
class DuplicateKeyException(PersistentOrderedDictException):
    pass
class PersistentOrderedDictIntegrityException(PersistentOrderedDictException):
    pass


def convert_to_tastypod(inpath, outpath):
    """
    Reads a t4k.PersistentOrderedDict from inpath and converts it to 
    tastypy.PersistentOrderedDict, stored at outpath.
    """
    t4k_pod = PersistentOrderedDict(inpath)
    tasty_pod = tastypy.POD(outpath)
    for key, val in t4k_pod:
        tasty_pod[key] = val


class PersistentOrderedDict(object):

    ESCAPE_TAB_PATTERN = re.compile('\t')
    UNESCAPE_TAB_PATTERN = re.compile(r'(?P<prefix>^|[^\\])\\t')
    ESCAPE_SLASH_PATTERN = re.compile(r'\\')
    UNESCAPE_SLASH_PATTERN = re.compile(r'\\\\')


    def __init__(
        self, 
        path,
        gzipped=False,
        lines_per_file=LINES_PER_FILE,
        verbose=False
    ):

        self.path = path
        self.lines_per_file = lines_per_file
        self.gzipped = gzipped
        self.verbose = verbose

        self.set_write_method(gzipped)
        self.ensure_path(path)

        # read in all data (if any)
        self.read()

        # Hold is off by default -> updates are immediately written to file
        self._hold = False

        # Keep track of files whose contents don't match values in memory
        self.dirty_files = set()

        # Always be sure to synchronize before the script exits
        atexit.register(self.unhold)


    def set_write_method(self, gzipped):

        if gzipped:
            self.open = gzip.open
        else:
            self.open = open


    def ensure_path(self, path):

        # if the path doesn't exist, make it
        if not os.path.exists(path):
            os.makedirs(path)

        # if the path exists but points to a file, raise
        elif os.path.isfile(path):
            raise IOError(
                'The path given to PersistentOrderedDict should correspond '
                'to a folder.  A file was found instead: %s' % path
            )


    def copy(self, path, lines_per_file, gzipped=False):

        self.gzipped = gzipped
        self.path = path
        self.lines_per_file = lines_per_file

        self.set_write_method(gzipped)
        self.ensure_path(path)

        num_files = int(math.ceil(
            len(self.data) / float(self.lines_per_file)
        ))
        self.dirty_files = set(range(num_files))
        self.maybe_sync()


    def hold(self):
        '''
        temporarily prevent writing updates to file
        '''
        self._hold = True

    
    def unhold(self):
        '''
        Resume writing updates to file, synchronize any dirty files.
        '''
        self._hold = False
        self.maybe_sync()


    def keys(self):
        return copy.copy(self.key_order)


    def values(self):
        return copy.copy([self.data[k] for k in self.key_order])


    def path_from_int(self, i):
        if self.gzipped:
            return os.path.abspath(os.path.join(self.path, '%d.json.gz' % i))
        else:
            return os.path.abspath(os.path.join(self.path, '%d.json' % i))


    def read(self):

        self.key_order = []
        self.index_lookup = {}
        self.data = {}

        i=0
        for fname in ls(self.path, dirs=False):

            fname = os.path.abspath(fname)

            if self.verbose:
                print fname

            # ensure that files are in expected order,
            # that none are missing, and that no lines are missing.
            if fname != self.path_from_int(i):
                raise PersistentOrderedDictIntegrityException(
                    'Expected %s but found %s.' 
                    % (self.path_from_int(i), fname)
                )

            if i > 0:
                prev_file_path = self.path_from_int(i-1)
                num_lines_prev_file = len(
                    self.open(prev_file_path, 'r').readlines()
                )
                if num_lines_prev_file != self.lines_per_file:
                    raise PersistentOrderedDictIntegrityException(
                        "PersistentOrderedDict: "
                        "A file on disk appears to be corrupted, because "
                        "it's missing lines: %s " % prev_file_path
                    )

            i += 1

            for entry in self.open(os.path.join(fname)):

                # skip blank lines (there's always one at end of file)
                if entry=='':
                    continue

                key, json_record = entry.split('\t', 1)
                key = self.UNESCAPE_TAB_PATTERN.sub('\g<prefix>\t', key)
                key = self.UNESCAPE_SLASH_PATTERN.sub(r'\\', key)
                key = key.decode('utf8')
                
                # remove the newline of the end of json_record, and read it
                record = json.loads(json_record[:-1])
                self.data[key] = record
                self.key_order.append(key)
                self.index_lookup[key] = len(self.key_order)-1


    def mark_dirty(self, key):

        key = self.ensure_unicode(key)
        index = self.index_lookup[key]
        file_num = index / self.lines_per_file
        self.dirty_files.add(file_num)


    def maybe_sync(self):
        """
        Write changes that have been made if self isn't under a "hold".
        """
        # To limit I/O consumption, sync'in is dactivated when hold is on.  
        if not self._hold:
            self.sync()


    def sync(self):
        """
        Write any changes that have been made to the datastructure to disk.
        """
        for file_num in self.dirty_files:

            # Get the dirty file
            try:
                path =  self.path_from_int(file_num)
            except TypeError:
                print file_num
                raise
            f = self.open(path, 'w')

            # Go through keys mapped to this file and re-write them
            start = file_num * self.lines_per_file
            stop = start + self.lines_per_file
            for key in self.key_order[start:stop]:

                record = self.data[key]
                # Escape tabs in key, and encode using utf8
                key = self.escape_key(key)
                f.write('%s\t%s\n' % (key, json.dumps(record)))

        # No more dirty files
        self.dirty_files = set()


    def escape_key(self, key):
        key = key.encode('utf8')
        key = self.ESCAPE_SLASH_PATTERN.sub(r'\\\\', key)
        key = self.ESCAPE_TAB_PATTERN.sub(r'\\t', key)
        return key


    def __iter__(self):
        self.pointer = 0
        return self


    def next(self):
        try:
            key = self.key_order[self.pointer]
        except IndexError:
            raise StopIteration

        val = self.data[key]
        self.pointer += 1

        return key, val


    def __contains__(self, key):
        key = self.ensure_unicode(key)
        return key in self.index_lookup


    def __len__(self):
        return len(self.key_order)


    def __getitem__(self, key):
        key = self.ensure_unicode(key)
        return self.data[key]


    def ensure_unicode(self, key):

        # ensure that the key is string-like
        if not isinstance(key, basestring):
            raise ValueError(
                'Keys must be str or unicode, and will be converted to '
                'unicode type internally.'
            )

        # cast the key into unicode if necessary
        if not isinstance(key, unicode):
            key = key.decode('utf8')

        return key


    def update(self, key):
        '''
        This can be called to ensure that a specific key will be 
        synchronized.  It's helpful if a mutable object is stored at that 
        key, since it could be changed without triggering __setitem__;
        this provides a way to notify PersistentOrderedDict that the value 
        at that key has changed.
        '''
        key = self.ensure_unicode(key)
        self.mark_dirty(key)
        self.maybe_sync()


    def set_item(self, key, val):

        key = self.ensure_unicode(key)
        val = copy.deepcopy(val)

        # if there isn't already an entry, we need to allocate a new slot
        if key not in self.data:
            self.key_order.append(key)
            self.index_lookup[key] = len(self.key_order)-1

        # update the value held at <key>
        self.data[key] = val
        self.mark_dirty(key)
        self.maybe_sync()


    def __setitem__(self, key, val):

        key = self.ensure_unicode(key)
        val = copy.deepcopy(val)

        # if there isn't already an entry, we need to allocate a new slot
        if key not in self.data:
            self.key_order.append(key)
            self.index_lookup[key] = len(self.key_order)-1

        # update the value held at <key>
        self.data[key] = val
        self.mark_dirty(key)
        self.maybe_sync()


    def set(self, key, subkey, value):
        key = self.ensure_unicode(key)
        self[key][subkey] = value
        self.update(key)


    def add(self, key):
        key = self.ensure_unicode(key)
        if key in self:
            raise DuplicateKeyException(
                'PersistentOrderedDict: key "%s" already exists.' % key)
        else:
            self[key] = True


    def add_if_absent(self, key):
        '''
        Same as add, but don't raise an error if the key exists, just do 
        nothing in that case.
        '''
        try:
            self.add(key)
        except DuplicateKeyException:
            pass


    def convert_to_tracker(self):

        # Rewrite every value to satisfy the form of a progress tracker
        self.hold()
        for key, spec in self:

            # Ensure the value at key is a dict, and add special keys
            if not isinstance(spec, dict):
                self[key] = {'val':spec, '_done':False, '_tries':0}
            else:
                self[key].update({'_done':False, '_tries':0})

            # Mark the key as updated
            self.update(key)

        self.unhold()


class ProgressTracker(PersistentOrderedDict):

    def check_or_add(self, key):
        '''
        checks if there is an entry for key already marked as done
        (returns True if so).  If no entry exists for key, it makes one
        and provides it with a defualt value of _done:False and _tries:0
        '''
        key = self.ensure_unicode(key)
        if key in self:
            if self[key]['_done']:
                return True
            else:
                return False
        else:
            self[key] = {'_done':False, '_tries':0}
            return False


    def check(self, key):
        key = self.ensure_unicode(key)
        if key in self:
            if self[key]['_done']:
                return True
            else:
                return False
        else:
            return False


    def add(self, key):
        key = self.ensure_unicode(key)
        if key in self:
            raise DuplicateKeyException(
                'ProgressTracker: key "%s" already exists.' % key)
        else:
            self[key] = {'_done':False, '_tries':0}


    def increment_tries(self, key):
        key = self.ensure_unicode(key)
        self[key]['_tries'] += 1
        self.update(key)


    def reset_tries(self, key):
        key = self.ensure_unicode(key)
        self[key]['_tries'] = 0
        self.update(key)


    def mark_done(self, key):
        key = self.ensure_unicode(key)
        self[key]['_done'] = True
        self.update(key)


    def mark_not_done(self, key):
        key = self.ensure_unicode(key)
        self[key]['_done'] = False
        self.update(key)



def requires_lock(lock):
    def decorator(f):
        def f_with_lock(*args, **kwargs):
            lock.acquire()
            return_val = f(*args, **kwargs)
            lock.release()
            return return_val

        return f_with_lock
    return decorator


def requires_tracker_open(f):
    def f_that_requires_tracker_open(self, *args, **kwargs):
        if not self.tracker_open:
            raise CalledClosedTrackerException
        return f(self, *args, **kwargs)

    return f_that_requires_tracker_open



class CalledClosedTrackerException(Exception):
    pass



class SharedProgressTracker(object):
    pass

    CLOSE = 0
    LOCK = multiprocessing.RLock()

    def __init__(self, path):

        self.client_pipe, server_pipe = multiprocessing.Pipe()
        self.lock = multiprocessing.Lock()

        # create a real progress_tracker and a listen loop around it
        client_tracker = multiprocessing.Process(
            target=progress_tracker_serve,
            args=(path, server_pipe)
        )
        client_tracker.start()

        self.tracker_open = True

    @requires_lock(LOCK)
    def close(self):
        self.client_pipe.send(self.CLOSE)

    @requires_tracker_open
    def hold(self):
        self.LOCK.acquire()
        self.client_pipe.send(('hold',))
        return self.client_pipe.recv()

    @requires_tracker_open
    def unhold(self):
        self.client_pipe.send(('unhold',))
        return_val = self.client_pipe.recv()
        self.LOCK.release()
        return return_val

    @requires_tracker_open
    @requires_lock(LOCK)
    def read(self):
        self.client_pipe.send(('read',))
        return self.client_pipe.recv()

    @requires_tracker_open
    @requires_lock(LOCK)
    def mark_dirty(self, key):
        self.client_pipe.send(('mark_dirty', key))
        return self.client_pipe.recv()

    @requires_tracker_open
    @requires_lock(LOCK)
    def sync(self):
        self.client_pipe.send(('sync',))
        return self.client_pipe.recv()

    @requires_tracker_open
    @requires_lock(LOCK)
    def __contains__(self, key):
        self.client_pipe.send(('__contains__', key))
        return self.client_pipe.recv()

    @requires_tracker_open
    @requires_lock(LOCK)
    def __len__(self):
        self.client_pipe.send(('__len__',))
        return self.client_pipe.recv()

    @requires_tracker_open
    @requires_lock(LOCK)
    def __getitem__(self, key):
        self.client_pipe.send(('__getitem__', key))
        return self.client_pipe.recv()

    @requires_tracker_open
    @requires_lock(LOCK)
    def update(self, key):
        self.client_pipe.send(('update', key))
        return self.client_pipe.recv()

    @requires_tracker_open
    @requires_lock(LOCK)
    def __setitem__(self, key, val):
        self.client_pipe.send(('__setitem__', key, val))
        return self.client_pipe.recv()

    @requires_tracker_open
    @requires_lock(LOCK)
    def check_or_add(self, key):
        self.client_pipe.send(('check_or_add', key))
        return self.client_pipe.recv()

    @requires_tracker_open
    @requires_lock(LOCK)
    def set(self, key, subkey, val):
        self.client_pipe.send(('set', key, subkey, val))
        return self.client_pipe.recv()

    @requires_tracker_open
    @requires_lock(LOCK)
    def check(self, key):
        self.client_pipe.send(('check', key))
        return self.client_pipe.recv()

    @requires_tracker_open
    @requires_lock(LOCK)
    def add(self, key):
        self.client_pipe.send(('add', key))
        return self.client_pipe.recv()

    @requires_tracker_open
    @requires_lock(LOCK)
    def increment_tries(self, key):
        self.client_pipe.send(('increment_tries', key))
        return self.client_pipe.recv()

    @requires_tracker_open
    @requires_lock(LOCK)
    def reset_tries(self, key):
        self.client_pipe.send(('reset_tries', key))
        return self.client_pipe.recv()

    @requires_tracker_open
    @requires_lock(LOCK)
    def mark_done(self, key):
        self.client_pipe.send(('mark_done', key))
        return self.client_pipe.recv()

    @requires_tracker_open
    @requires_lock(LOCK)
    def mark_not_done(self, key):
        self.client_pipe.send(('mark_not_done', key))
        return self.client_pipe.recv()


    
def progress_tracker_serve(path, pipe):
    progress_tracker = ProgressTracker(path)
    is_open = True
    while is_open:
        message = pipe.recv()
        if message == SharedProgressTracker.CLOSE:
            is_open = False
        else:
            attr, args = message[0], message[1:]
            pipe.send(getattr(progress_tracker, attr)(*args))

        

# This was used to ensure that the PersistentOrderedDict got the chance
# to synchronize with the filesystem before the script exited (and likewise
# for the ProgressTracker.  Now the atexit module is being used.
#
#class GracefulDeath(object):
#    '''
#    Catches kill singals and acknowledges kill signals so that other
#    clients can finish critical steps before dying
#
#    The client callable should instantiate a GracefulDeath, and then
#    watch it's kill_now flag to know when it should die
#    '''
#    kill_now = False
#    def __init__(self):
#        signal.signal(signal.SIGINT, self.exit_gracefully)
#        signal.signal(signal.SIGTERM, self.exit_gracefully)
#
#    def exit_gracefully(self,signum, frame):
#        self.kill_now = True



import re

url = re.compile('(?:\\b[a-z\\d.-]+://[^<>\\s]+|\\b(?:(?:(?:[^\\s!@#$%^&*()_=+[\\]{}\\|;:\\\'",.<>/?]+)\\.)+(?:ac|ad|aero|ae|af|ag|ai|al|am|an|ao|aq|arpa|ar|asia|as|at|au|aw|ax|az|ba|bb|bd|be|bf|bg|bh|biz|bi|bj|bm|bn|bo|br|bs|bt|bv|bw|by|bz|cat|ca|cc|cd|cf|cg|ch|ci|ck|cl|cm|cn|coop|com|co|cr|cu|cv|cx|cy|cz|de|dj|dk|dm|do|dz|ec|edu|ee|eg|er|es|et|eu|fi|fj|fk|fm|fo|fr|ga|gb|gd|ge|gf|gg|gh|gi|gl|gm|gn|gov|gp|gq|gr|gs|gt|gu|gw|gy|hk|hm|hn|hr|ht|hu|id|ie|il|im|info|int|in|io|iq|ir|is|it|je|jm|jobs|jo|jp|ke|kg|kh|ki|km|kn|kp|kr|kw|ky|kz|la|lb|lc|li|lk|lr|ls|lt|lu|lv|ly|ma|mc|md|me|mg|mh|mil|mk|ml|mm|mn|mobi|mo|mp|mq|mr|ms|mt|museum|mu|mv|mw|mx|my|mz|name|na|nc|net|ne|nf|ng|ni|nl|no|np|nr|nu|nz|om|org|pa|pe|pf|pg|ph|pk|pl|pm|pn|pro|pr|ps|pt|pw|py|qa|re|ro|rs|ru|rw|sa|sb|sc|sd|se|sg|sh|si|sj|sk|sl|sm|sn|so|sr|st|su|sv|sy|sz|tc|td|tel|tf|tg|th|tj|tk|tl|tm|tn|to|tp|travel|tr|tt|tv|tw|tz|ua|ug|uk|um|us|uy|uz|va|vc|ve|vg|vi|vn|vu|wf|ws|xn--0zwm56d|xn--11b5bs3a9aj6g|xn--80akhbyknj4f|xn--9t4b11yi5a|xn--deba0ad|xn--g6w251d|xn--hgbk6aj7f53bba|xn--hlcj6aya9esc7a|xn--jxalpdlp|xn--kgbechtv|xn--zckzah|ye|yt|yu|za|zm|zw)|(?:(?:[0-9]|[1-9]\\d|1\\d{2}|2[0-4]\\d|25[0-5])\\.){3}(?:[0-9]|[1-9]\\d|1\\d{2}|2[0-4]\\d|25[0-5]))(?:[;/][^#?<>\\s]*)?(?:\\?[^#<>\\s]*)?(?:#[^<>\\s]*)?(?!\\w))')
url_with_scheme = re.compile('(?:\\b[a-z\\d.-]+://[^<>\\s]+|\\b(?:(?:(?:[^\\s!@#$%^&*()_=+[\\]{}\\|;:\\\'",.<>/?]+)\\.)+(?:ac|ad|aero|ae|af|ag|ai|al|am|an|ao|aq|arpa|ar|asia|as|at|au|aw|ax|az|ba|bb|bd|be|bf|bg|bh|biz|bi|bj|bm|bn|bo|br|bs|bt|bv|bw|by|bz|cat|ca|cc|cd|cf|cg|ch|ci|ck|cl|cm|cn|coop|com|co|cr|cu|cv|cx|cy|cz|de|dj|dk|dm|do|dz|ec|edu|ee|eg|er|es|et|eu|fi|fj|fk|fm|fo|fr|ga|gb|gd|ge|gf|gg|gh|gi|gl|gm|gn|gov|gp|gq|gr|gs|gt|gu|gw|gy|hk|hm|hn|hr|ht|hu|id|ie|il|im|info|int|in|io|iq|ir|is|it|je|jm|jobs|jo|jp|ke|kg|kh|ki|km|kn|kp|kr|kw|ky|kz|la|lb|lc|li|lk|lr|ls|lt|lu|lv|ly|ma|mc|md|me|mg|mh|mil|mk|ml|mm|mn|mobi|mo|mp|mq|mr|ms|mt|museum|mu|mv|mw|mx|my|mz|name|na|nc|net|ne|nf|ng|ni|nl|no|np|nr|nu|nz|om|org|pa|pe|pf|pg|ph|pk|pl|pm|pn|pro|pr|ps|pt|pw|py|qa|re|ro|rs|ru|rw|sa|sb|sc|sd|se|sg|sh|si|sj|sk|sl|sm|sn|so|sr|st|su|sv|sy|sz|tc|td|tel|tf|tg|th|tj|tk|tl|tm|tn|to|tp|travel|tr|tt|tv|tw|tz|ua|ug|uk|um|us|uy|uz|va|vc|ve|vg|vi|vn|vu|wf|ws|xn--0zwm56d|xn--11b5bs3a9aj6g|xn--80akhbyknj4f|xn--9t4b11yi5a|xn--deba0ad|xn--g6w251d|xn--hgbk6aj7f53bba|xn--hlcj6aya9esc7a|xn--jxalpdlp|xn--kgbechtv|xn--zckzah|ye|yt|yu|za|zm|zw)|(?:(?:[0-9]|[1-9]\\d|1\\d{2}|2[0-4]\\d|25[0-5])\\.){3}(?:[0-9]|[1-9]\\d|1\\d{2}|2[0-4]\\d|25[0-5]))(?:[;/][^#?<>\\s]*)?(?:\\?[^#<>\\s]*)?(?:#[^<>\\s]*)?(?!\\w))')
